#!/usr/bin/python
#coding:utf-8
#
#   PROGRAM/MODULE: firebird-base
#   FILE:           test/test_buffer.py
#   DESCRIPTION:    Unit tests for firebird.base.buffer
#   CREATED:        14.5.2020
#
#  Software distributed under the License is distributed AS IS,
#  WITHOUT WARRANTY OF ANY KIND, either express or implied.
#  See the License for the specific language governing rights
#  and limitations under the License.
#
#  The Original Code was created by Pavel Cisar
#
#  Copyright (c) Pavel Cisar <pcisar@users.sourceforge.net>
#  and all contributors signed below.
#
#  All Rights Reserved.
#  Contributor(s): Pavel Císař (original code)
#                  ______________________________________.
#
# See LICENSE.TXT for details.

from __future__ import annotations
import unittest
from firebird.base.buffer import *

class TestBuffer(unittest.TestCase):
    """Unit tests for firebird.base.buffer with BytesBufferFactory"""
    def __init__(self, methodName='runTest'):
        super().__init__(methodName)
        self.factory = BytesBufferFactory
    def setUp(self) -> None:
        pass
    def tearDown(self):
        pass
    def assertBuffer(self, buffer, content):
        self.assertEqual(buffer.raw, content)
    def test_create(self):
        # Empty buffer
        buf = MemoryBuffer(0, factory=self.factory)
        self.assertEqual(buf.pos, 0)
        self.assertEqual(len(buf.raw), 0)
        self.assertIsNone(buf.eof_marker)
        self.assertIs(buf.max_size, UNLIMITED)
        self.assertIs(buf.byteorder, ByteOrder.LITTLE)
        self.assertTrue(buf.is_eof())
        self.assertEqual(buf.buffer_size, 0)
        self.assertEqual(buf.last_data, -1)
        # Sized
        buf = MemoryBuffer(10, factory=self.factory)
        self.assertEqual(buf.pos, 0)
        self.assertEqual(len(buf.raw), 10)
        self.assertBuffer(buf, b'\x00' * 10)
        #self.assertEqual(buf.raw, b'\x00' * 10)
        self.assertIsNone(buf.eof_marker)
        self.assertIs(buf.max_size, UNLIMITED)
        self.assertIs(buf.byteorder, ByteOrder.LITTLE)
        self.assertFalse(buf.is_eof())
        self.assertEqual(buf.buffer_size, 10)
        self.assertEqual(buf.last_data, -1)
        # Initialized
        buf = MemoryBuffer(b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x00\x00', factory=self.factory)
        self.assertEqual(buf.pos, 0)
        self.assertEqual(len(buf.raw), 12)
        self.assertBuffer(buf, b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x00\x00')
        self.assertIsNone(buf.eof_marker)
        self.assertIs(buf.max_size, UNLIMITED)
        self.assertIs(buf.byteorder, ByteOrder.LITTLE)
        self.assertFalse(buf.is_eof())
        self.assertEqual(buf.buffer_size, 12)
        self.assertEqual(buf.last_data, 9)
        # Max. Size
        buf = MemoryBuffer(10, max_size=20, factory=self.factory)
        self.assertEqual(buf.buffer_size, 10)
        self.assertIs(buf.max_size, 20)
        # Byte order
        buf = MemoryBuffer(10, byteorder=ByteOrder.BIG, factory=self.factory)
        self.assertIs(buf.byteorder, ByteOrder.BIG)
    def test_clear(self):
        # Empty
        buf = MemoryBuffer(0, factory=self.factory)
        buf.write(b'0123456789')
        buf.clear()
        self.assertEqual(buf.pos, 0)
        self.assertEqual(len(buf.raw), 10)
        self.assertBuffer(buf, b'\x00' * 10)
        self.assertFalse(buf.is_eof())
        self.assertEqual(buf.buffer_size, 10)
        self.assertEqual(buf.last_data, -1)
        # Sized
        buf = MemoryBuffer(10, factory=self.factory)
        for i in range(buf.buffer_size):
            buf.raw[i] = 255
        self.assertBuffer(buf, b'\xff' * 10)
        buf.clear()
        self.assertEqual(buf.pos, 0)
        self.assertEqual(len(buf.raw), 10)
        self.assertBuffer(buf, b'\x00' * 10)
        self.assertFalse(buf.is_eof())
        self.assertEqual(buf.buffer_size, 10)
        self.assertEqual(buf.last_data, -1)
    def test_write(self):
        buf = MemoryBuffer(0, factory=self.factory)
        buf.write(b'ABCDE')
        self.assertEqual(buf.pos, 5)
        self.assertBuffer(buf, b'ABCDE')
        self.assertTrue(buf.is_eof())
    def test_write_byte(self):
        buf = MemoryBuffer(0, factory=self.factory)
        buf.write_byte(1)
        self.assertEqual(buf.pos, 1)
        self.assertBuffer(buf, b'\x01')
        self.assertTrue(buf.is_eof())
    def test_write_short(self):
        buf = MemoryBuffer(0, factory=self.factory)
        buf.write_short(2)
        self.assertEqual(buf.pos, 2)
        self.assertBuffer(buf, b'\x02\x00')
        self.assertTrue(buf.is_eof())
    def test_write_int(self):
        buf = MemoryBuffer(0, factory=self.factory)
        buf.write_int(3)
        self.assertEqual(buf.pos, 4)
        self.assertBuffer(buf, b'\x03\x00\x00\x00')
        self.assertTrue(buf.is_eof())
    def test_write_bigint(self):
        buf = MemoryBuffer(0, factory=self.factory)
        buf.write_bigint(4)
        self.assertEqual(buf.pos, 8)
        self.assertBuffer(buf, b'\x04\x00\x00\x00\x00\x00\x00\x00')
        self.assertTrue(buf.is_eof())
    def test_write_number(self):
        buf = MemoryBuffer(0, factory=self.factory)
        buf.write_number(255, 1)
        self.assertEqual(buf.pos, 1)
        self.assertBuffer(buf, b'\xff')
        self.assertTrue(buf.is_eof())
        #
        buf = MemoryBuffer(0, factory=self.factory)
        buf.write_number(255, 2)
        self.assertEqual(buf.pos, 2)
        self.assertBuffer(buf, b'\xff\x00')
        self.assertTrue(buf.is_eof())
        #
        buf = MemoryBuffer(0, factory=self.factory)
        buf.write_number(255, 4)
        self.assertEqual(buf.pos, 4)
        self.assertBuffer(buf, b'\xff\x00\x00\x00')
        self.assertTrue(buf.is_eof())
        #
        buf = MemoryBuffer(0, factory=self.factory)
        buf.write_number(255, 8)
        self.assertEqual(buf.pos, 8)
        self.assertBuffer(buf, b'\xff\x00\x00\x00\x00\x00\x00\x00')
        self.assertTrue(buf.is_eof())
        # Atypical sizes
        buf = MemoryBuffer(0, factory=self.factory)
        buf.write_number(255, 3)
        self.assertEqual(buf.pos, 3)
        self.assertBuffer(buf, b'\xff\x00\x00')
        self.assertTrue(buf.is_eof())
        buf = MemoryBuffer(0, factory=self.factory)
        buf.write_number(255, 12)
        self.assertEqual(buf.pos, 12)
        self.assertBuffer(buf, b'\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')
        self.assertTrue(buf.is_eof())
    def test_write_string(self):
        buf = MemoryBuffer(0, factory=self.factory)
        buf.write_string('string')
        self.assertEqual(buf.pos, 7)
        self.assertBuffer(buf, b'string\x00')
        self.assertTrue(buf.is_eof())
    def test_write_pascal_string(self):
        buf = MemoryBuffer(0, factory=self.factory)
        buf.write_pascal_string('string')
        self.assertEqual(buf.pos, 7)
        self.assertBuffer(buf, b'\x06string')
        self.assertTrue(buf.is_eof())
    def test_write_past_size(self):
        buf = MemoryBuffer(0, max_size=5, factory=self.factory)
        buf.write(b'ABCDE')
        with self.assertRaises(IOError) as cm:
            buf.write(b'exceeds size')
        self.assertEqual(cm.exception.args, ("Cannot resize buffer past max. size 5 bytes",))
    def test_read(self):
        buf = MemoryBuffer(b'ABCDE', factory=self.factory)
        self.assertEqual(buf.read(3), b'ABC')
        self.assertEqual(buf.pos, 3)
        self.assertFalse(buf.is_eof())
        self.assertEqual(buf.read(), b'DE')
        self.assertEqual(buf.pos, 5)
        self.assertTrue(buf.is_eof())
    def test_read_byte(self):
        buf = MemoryBuffer(b'\x01', factory=self.factory)
        self.assertEqual(buf.read_byte(), 1)
        self.assertEqual(buf.pos, 1)
        self.assertTrue(buf.is_eof())
    def test_read_short(self):
        buf = MemoryBuffer(b'\x02\x00', factory=self.factory)
        self.assertEqual(buf.read_short(), 2)
        self.assertEqual(buf.pos, 2)
        self.assertTrue(buf.is_eof())
    def test_read_int(self):
        buf = MemoryBuffer(b'\x03\x00\x00\x00', factory=self.factory)
        self.assertEqual(buf.read_int(), 3)
        self.assertEqual(buf.pos, 4)
        self.assertTrue(buf.is_eof())
    def test_read_bigint(self):
        buf = MemoryBuffer(b'\x04\x00\x00\x00\x00\x00\x00\x00', factory=self.factory)
        self.assertEqual(buf.read_bigint(), 4)
        self.assertEqual(buf.pos, 8)
        self.assertBuffer(buf, b'\x04\x00\x00\x00\x00\x00\x00\x00')
        self.assertTrue(buf.is_eof())
    def test_read_number(self):
        buf = MemoryBuffer(b'\xff', factory=self.factory)
        self.assertEqual(buf.read_number(1), 255)
        self.assertEqual(buf.pos, 1)
        self.assertTrue(buf.is_eof())
        #
        buf = MemoryBuffer(b'\xff\x00', factory=self.factory)
        self.assertEqual(buf.read_number(2), 255)
        self.assertEqual(buf.pos, 2)
        self.assertTrue(buf.is_eof())
        #
        buf = MemoryBuffer(b'\xff\x00\x00\x00', factory=self.factory)
        self.assertEqual(buf.read_number(4), 255)
        self.assertEqual(buf.pos, 4)
        self.assertTrue(buf.is_eof())
        #
        buf = MemoryBuffer(b'\xff\x00\x00\x00\x00\x00\x00\x00', factory=self.factory)
        self.assertEqual(buf.read_number(8), 255)
        self.assertEqual(buf.pos, 8)
        self.assertTrue(buf.is_eof())
        # Atypical sizes
        buf = MemoryBuffer(b'\xff\x00\x00', factory=self.factory)
        self.assertEqual(buf.read_number(3), 255)
        self.assertEqual(buf.pos, 3)
        self.assertTrue(buf.is_eof())
        buf = MemoryBuffer(b'\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00', factory=self.factory)
        self.assertEqual(buf.read_number(12), 255)
        self.assertEqual(buf.pos, 12)
        self.assertTrue(buf.is_eof())
    def test_read_sized_int(self):
        buf = MemoryBuffer(b'\x04\x00\x03\x00\x00\x00', factory=self.factory)
        self.assertEqual(buf.read_sized_int(), 3)
        self.assertEqual(buf.pos, 6)
        self.assertTrue(buf.is_eof())
    def test_read_string(self):
        buf = MemoryBuffer(b'string 1\x00string 2\x00', factory=self.factory)
        self.assertEqual(buf.read_string(), 'string 1')
        self.assertEqual(buf.pos, 9)
        self.assertFalse(buf.is_eof())
        #  No zero-terminator
        buf = MemoryBuffer(b'string', factory=self.factory)
        self.assertEqual(buf.read_string(), 'string')
        self.assertEqual(buf.pos, 7)
        self.assertTrue(buf.is_eof())
    def test_read_pascal_string(self):
        buf = MemoryBuffer(b'\x06string', factory=self.factory)
        self.assertEqual(buf.read_pascal_string(), 'string')
        self.assertEqual(buf.pos, 7)
        self.assertTrue(buf.is_eof())
    def test_read_sized_string(self):
        buf = MemoryBuffer(b'\x08\x00string 1\x08\x00string 2', factory=self.factory)
        self.assertEqual(buf.read_sized_string(), 'string 1')
        self.assertEqual(buf.pos, 10)
        self.assertFalse(buf.is_eof())
    def test_read_bytes(self):
        buf = MemoryBuffer(b'\x08\x00ABCDEFGH\x08\x00string 2', factory=self.factory)
        self.assertEqual(buf.read_bytes(), b'ABCDEFGH')
        self.assertEqual(buf.pos, 10)
        self.assertFalse(buf.is_eof())
    def test_read_past_size(self):
        buf = MemoryBuffer(b'ABCDE', factory=self.factory)
        with self.assertRaises(IOError) as cm:
            buf.read_bigint()
        self.assertEqual(cm.exception.args, ("Insufficient buffer size",))
    def test_eof_marker(self):
        buf = MemoryBuffer(b'\x08\x00ABCDEFGH\xFF\x00\x00\x00\x00\x00\x00', eof_marker=255,
                           factory=self.factory)
        while not buf.is_eof():
            buf.pos += 1
        self.assertLess(buf.pos, buf.buffer_size)
        self.assertEqual(buf.pos, 10)
        self.assertEqual(safe_ord(buf.raw[buf.pos]), buf.eof_marker)

class TestCBuffer(TestBuffer):
    """Unit tests for firebird.base.buffer with CTypesBufferFactory"""
    def __init__(self, methodName='runTest'):
        super().__init__(methodName)
        self.factory = CTypesBufferFactory
    def assertBuffer(self, buffer, content):
        self.assertEqual(buffer.raw.raw, content)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
#coding:utf-8
#
#   PROGRAM/MODULE: firebird-base
#   FILE:           test/test_types.py
#   DESCRIPTION:    Unit tests for firebird.base.types
#   CREATED:        14.5.2020
#
#  Software distributed under the License is distributed AS IS,
#  WITHOUT WARRANTY OF ANY KIND, either express or implied.
#  See the License for the specific language governing rights
#  and limitations under the License.
#
#  The Original Code was created by Pavel Cisar
#
#  Copyright (c) Pavel Cisar <pcisar@users.sourceforge.net>
#  and all contributors signed below.
#
#  All Rights Reserved.
#  Contributor(s): Pavel Císař (original code)
#                  ______________________________________.
#
# See LICENSE.TXT for details.

from __future__ import annotations
from typing import List
import unittest
from dataclasses import dataclass
from firebird.base.types import *

class TestTypes(unittest.TestCase):
    """Unit tests for firebird.base.types"""
    def __init__(self, methodName='runTest'):
        super().__init__(methodName)
        self.output: List = []
    def setUp(self) -> None:
        self.output.clear()
    def tearDown(self):
        pass
    def test_exceptions(self):
        "Test exceptions"
        e = Error("Message", code=1, subject=self)
        self.assertTupleEqual(e.args, ("Message",))
        self.assertEqual(e.code, 1)
        self.assertEqual(e.subject, self)
        self.assertIsNone(e.other_attr)
    def test_singletons(self):
        "Test Singletons"
        class MySingleton(Singleton):
            pass

        class MyOtherSingleton(MySingleton):
            pass
        #
        s = MySingleton()
        self.assertIs(s, MySingleton())
        os = MyOtherSingleton()
        self.assertIs(os, MyOtherSingleton())
        self.assertIsNot(s, os)
    def test_sentinel(self):
        "Test Sentinel"
        self.assertEqual(UNKNOWN.name, 'UNKNOWN')
        self.assertEqual(str(UNKNOWN), 'UNKNOWN')
        self.assertEqual(repr(UNKNOWN), "Sentinel('UNKNOWN')")
        self.assertDictEqual(UNKNOWN.instances, {'DEFAULT': DEFAULT,
                                                 'INFINITY': INFINITY,
                                                 'UNLIMITED': UNLIMITED,
                                                 'UNKNOWN': UNKNOWN,
                                                 'NOT_FOUND': NOT_FOUND,
                                                 'UNDEFINED': UNDEFINED,
                                                 'ANY': ANY,
                                                 'ALL': ALL,
                                                 'SUSPEND': SUSPEND,
                                                 'RESUME': RESUME,
                                                 'STOP': STOP,
                                                 })
        for name, sentinel in Sentinel.instances.items():
            self.assertEqual(sentinel, Sentinel(name))
        self.assertNotIn('TEST-SENTINEL', Sentinel.instances)
        Sentinel('TEST-SENTINEL')
        self.assertIn('TEST-SENTINEL', Sentinel.instances)
    def test_distinct(self):
        "Test Distinct"
        @dataclass
        class MyDistinct(Distinct):
            key_1: int
            key_2: str
            payload: str
            def get_key(self):
                if not hasattr(self, '__key__'):
                    self.__key__ = (self.key_1, self.key_2)
                return self.__key__

        d = MyDistinct(1, 'A', '1A')
        self.assertFalse(hasattr(d, '__key__'))
        self.assertEqual(d.get_key(), (1, 'A'))
        self.assertTrue(hasattr(d, '__key__'))
        d.key_2 = 'B'
        self.assertEqual(d.get_key(), (1, 'A'))
    def test_cached_distinct(self):
        "Test CachedDistinct"
        class MyCachedDistinct(CachedDistinct):
            def __init__(self, key_1, key_2, payload):
                self.key_1 = key_1
                self.key_2 = key_2
                self.payload = payload
            @classmethod
            def extract_key(cls, *args, **kwargs) -> t.Hashable:
                return (args[0], args[1])
            def get_key(self) -> t.Hashable:
                return (self.key_1, self.key_2)
        #
        self.assertTrue(hasattr(MyCachedDistinct, '_instances_'))
        cd_1 = MyCachedDistinct(1, ANY, 'type 1A')
        self.assertIs(cd_1, MyCachedDistinct(1, ANY, 'type 1A'))
        self.assertIsNot(cd_1, MyCachedDistinct(2, ANY, 'type 2A'))
        self.assertTrue(hasattr(MyCachedDistinct, '_instances_'))
        self.assertEqual(len(getattr(MyCachedDistinct, '_instances_')), 1)
        cd_2 = MyCachedDistinct(2, ANY, 'type 2A')
        self.assertEqual(len(getattr(MyCachedDistinct, '_instances_')), 2)
        temp = MyCachedDistinct(2, ANY, 'type 2A')
        self.assertEqual(len(getattr(MyCachedDistinct, '_instances_')), 2)
        del cd_1, cd_2, temp
        self.assertEqual(len(getattr(MyCachedDistinct, '_instances_')), 0)
    def test_zmqaddress(self):
        "Test ZMQAddress"
        addr = ZMQAddress('ipc://@my-address')
        self.assertEqual(addr.address, '@my-address')
        self.assertEqual(addr.protocol, ZMQTransport.IPC)
        self.assertEqual(addr.domain, ZMQDomain.NODE)
        #
        addr = ZMQAddress('inproc://my-address')
        self.assertEqual(addr.address, 'my-address')
        self.assertEqual(addr.protocol, ZMQTransport.INPROC)
        self.assertEqual(addr.domain, ZMQDomain.LOCAL)
        #
        addr = ZMQAddress('tcp://127.0.0.1:*')
        self.assertEqual(addr.address, '127.0.0.1:*')
        self.assertEqual(addr.protocol, ZMQTransport.TCP)
        self.assertEqual(addr.domain, ZMQDomain.NODE)
        #
        addr = ZMQAddress('tcp://192.168.0.1:8001')
        self.assertEqual(addr.address, '192.168.0.1:8001')
        self.assertEqual(addr.protocol, ZMQTransport.TCP)
        self.assertEqual(addr.domain, ZMQDomain.NETWORK)
        #
        addr = ZMQAddress('pgm://192.168.0.1:8001')
        self.assertEqual(addr.address, '192.168.0.1:8001')
        self.assertEqual(addr.protocol, ZMQTransport.PGM)
        self.assertEqual(addr.domain, ZMQDomain.NETWORK)
        # Bytes
        addr = ZMQAddress(b'ipc://@my-address')
        self.assertEqual(addr.address, '@my-address')
        self.assertEqual(addr.protocol, ZMQTransport.IPC)
        self.assertEqual(addr.domain, ZMQDomain.NODE)
        # Bad ZMQ address
        with self.assertRaises(ValueError) as cm:
            addr = ZMQAddress('onion://@my-address')
        self.assertEqual(cm.exception.args, ("Unknown protocol 'onion'",))
        with self.assertRaises(ValueError) as cm:
            addr = ZMQAddress('192.168.0.1:8001')
        self.assertEqual(cm.exception.args, ("Protocol specification required",))
        with self.assertRaises(ValueError) as cm:
            addr = ZMQAddress('unknown://192.168.0.1:8001')
        self.assertEqual(cm.exception.args, ("Invalid protocol",))


if __name__ == '__main__':
    unittest.main()



#coding:utf-8
#
# PROGRAM/MODULE: firebird-base
# FILE:           firebird/base/types.py
# DESCRIPTION:    Types
# CREATED:        14.5.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Firebird Base - Types
"""

from __future__ import annotations
from typing import Any, Dict, Hashable, Callable, AnyStr, cast
from abc import ABC, ABCMeta, abstractmethod
from enum import Enum, IntEnum
from weakref import WeakValueDictionary

# Exceptions

class Error(Exception):
    """Exception that is intended to be used as a base class of all **application-related**
errors. The important difference from `Exception` class is that `Error` accepts keyword
arguments, that are stored into instance attributes with the same name.

Important:
    Attribute lookup on this class never fails, as all attributes that are not actually set,
    have `None` value.

Example:
    .. code-block:: python

        try:
            if condition:
                raise Error("Error message", err_code=1)
            else:
                raise Error("Unknown error")
        except Error as e:
            if e.err_code is None:
                ...
            elif e.err_code == 1:
                ...

Note:
    Warnings are not considered errors and thus should not use this class as base.
"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args)
        for name, value in kwargs.items():
            setattr(self, name, value)
    def __getattr__(self, name):
        return None

# Singletons

_singletons_ = {}

class SingletonMeta(type):
    """Metaclass for `Singleton` classes.

Manages internal cache of class instances. If instance for a class is in cache, it's returned
without calling the constructor, otherwise the instance is created normally and stored in
cache for later use.
"""
    def __call__(cls: Singleton, *args, **kwargs):
        name = f"{cls.__module__}.{cls.__qualname__}"
        obj = _singletons_.get(name)
        if obj is None:
            obj = super().__call__(*args, **kwargs)
            _singletons_[name] = obj
        return obj

class Singleton(metaclass=SingletonMeta):
    """Base class for singletons.

Important:
    If you create a descendant class that uses constructor arguments, these arguments are
    meaningful ONLY on first call, because all subsequent calls simply return an instance
    stored in cache without calling the constructor.
"""

# Sentinels

class SentinelMeta(type):
    "Metaclass for `Sentinel`."
    def __call__(cls: Sentinel, *args, **kwargs):
        name = args[0].upper()
        obj = cls.instances.get(name)
        if obj is None:
            obj = super().__call__(*args, **kwargs)
            cls.instances[name] = obj
        return obj

class Sentinel(metaclass=SentinelMeta):
    """Simple sentinel object.

Important:
  All sentinels have name, that is **always in capital letters**. Sentinels with the same
  name are singletons.

Attributes:
    name (str): Sentinel name.
"""
    #: Class attribute with defined sentinels. There is no need to access or manipulate it.
    instances = {}
    def __init__(self, name: str):
        self.name = name.upper()
    def __str__(self):
        "Returns name"
        return self.name
    def __repr__(self):
        "Returns Sentinel('name')"
        return f"Sentinel('{self.name}')"

# Useful sentinel objects

#: Sentinel that denotes default value
DEFAULT: Sentinel = Sentinel('DEFAULT')
#: Sentinel that denotes infinity value
INFINITY: Sentinel = Sentinel('INFINITY')
#: Sentinel that denotes unlimited value
UNLIMITED: Sentinel = Sentinel('UNLIMITED')
#: Sentinel that denotes unknown value
UNKNOWN: Sentinel = Sentinel('UNKNOWN')
#: Sentinel that denotes a condition when value was not found
NOT_FOUND: Sentinel = Sentinel('NOT_FOUND')
#: Sentinel that denotes explicitly undefined value
UNDEFINED: Sentinel = Sentinel('UNDEFINED')
#: Sentinel that denotes any value
ANY: Sentinel = Sentinel('ANY')
#: Sentinel that denotes all possible values
ALL: Sentinel = Sentinel('ALL')
#: Sentinel that denotes suspend request (in message queue)
SUSPEND: Sentinel = Sentinel('SUSPEND')
#: Sentinel that denotes resume request (in message queue)
RESUME: Sentinel = Sentinel('RESUME')
#: Sentinel that denotes stop request (in message queue)
STOP: Sentinel = Sentinel('STOP')

# Distinct objects

class Distinct(ABC):
    """Abstract base class for classes (incl. dataclasses) with distinct instances.
"""
    @abstractmethod
    def get_key(self) -> Hashable:
        """Returns instance key.

Important:
    The key is used for instance hash computation that by default uses the `hash`
    function. If the key is not suitable argument for `hash`, you must provide your
    own `__hash__` implementation as well!
"""
    __hash__ = lambda self: hash(self.get_key())

class CachedDistinctMeta(ABCMeta):
    "Metaclass for CachedDistinct."
    def __call__(cls: CachedDistinct, *args, **kwargs):
        key = cls.extract_key(*args, **kwargs)
        obj = cls._instances_.get(key)
        if obj is None:
            obj = super().__call__(*args, **kwargs)
            cls._instances_[key] = obj
        return obj

class CachedDistinct(Distinct, metaclass=CachedDistinctMeta):
    """Abstract `Distinct` descendant that caches instances.

All created instances are cached in `~weakref.WeakValueDictionary`.
"""
    def __init_subclass__(cls: Type, /, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        setattr(cls, '_instances_', WeakValueDictionary())
    @classmethod
    @abstractmethod
    def extract_key(cls, *args, **kwargs) -> Hashable:
        """Returns key from arguments passed to `__init__()`.

Important:
    The key is used to store instance in cache. It should be the same as key returned by
    instance `Distinct.get_key()`!
"""

# Enums

class ByteOrder(Enum):
    "Byte order for storing numbers in binary `.MemoryBuffer`."
    LITTLE = 'little'
    BIG = 'big'
    NETWORK = BIG

class ZMQTransport(IntEnum):
    """ZeroMQ transport protocol"""
    UNKNOWN = 0 # Not a valid option, defined only to handle undefined values
    INPROC = 1
    IPC = 2
    TCP = 3
    PGM = 4
    EPGM = 5
    VMCI = 6

class ZMQDomain(IntEnum):
    """ZeroMQ address domain"""
    UNKNOWN = 0  # Not a valid option, defined only to handle undefined values
    LOCAL = 1    # Within process (inproc)
    NODE = 2     # On single node (ipc or tcp loopback)
    NETWORK = 3  # Network-wide (ip address or domain name)

# Enhanced string types

class ZMQAddress(str):
    """ZeroMQ endpoint address.

It behaves like `str`, but checks that value is valid ZMQ endpoint address, has
additional R/O properties and meaningful `repr()`.

Raises:
    ValueError: When string value passed to constructor is not a valid ZMQ endpoint address.
"""
    def __new__(cls, value: AnyStr):
        if isinstance(value, bytes):
            value = cast(bytes, value).decode('utf8')
        if '://' in value:
            protocol, _ = value.split('://', 1)
            if protocol.upper() not in ZMQTransport._member_map_:
                raise ValueError(f"Unknown protocol '{protocol}'")
            if protocol.upper() == 'UNKNOWN':
                raise ValueError("Invalid protocol")
        else:
            raise ValueError("Protocol specification required")
        return str.__new__(cls, value.lower())
    def __repr__(self):
        return f"ZMQAddress('{self}')"
    @property
    def protocol(self) -> ZMQTransport:
        "Transport protocol"
        protocol, _ = self.split('://', 1)
        return ZMQTransport._member_map_[protocol.upper()]
    @property
    def address(self) -> str:
        "Address"
        _, address = self.split('://', 1)
        return address
    @property
    def domain(self) -> ZMQDomain:
        "Address domain"
        if self.protocol == ZMQTransport.INPROC:
            return ZMQDomain.LOCAL
        if self.protocol == ZMQTransport.IPC:
            return ZMQDomain.NODE
        if self.protocol == ZMQTransport.TCP:
            if self.address.startswith('127.0.0.1') or self.address.lower().startswith('localhost'):
                return ZMQDomain.NODE
            return ZMQDomain.NETWORK
        # PGM, EPGM and VMCI
        return ZMQDomain.NETWORK

class MIME(str):
    """MIME type specification.

It behaves like `str`, but checks that value is valid MIME type specification, has
additional R/O properties and meaningful `repr()`.

"""
    #: Supported MIME types
    MIME_TYPES = ['text', 'image', 'audio', 'video', 'application', 'multipart', 'message']
    def __new__(cls, value: AnyStr):
        dfm = [x for x in value.split(';')]
        mime_type: str = dfm.pop(0)
        if (i := mime_type.find('/')) == -1:
            raise ValueError("MIME type specification must be 'type/subtype[;param=value;...]'")
        if mime_type[:i] not in cls.MIME_TYPES:
            raise ValueError(f"MIME type '{mime_type[:i]}' not supported")
        if [i for i in dfm if '=' not in i]:
            raise ValueError("Wrong specification of MIME type parameters")
        return str.__new__(cls, value)
    def __repr__(self):
        return f"MIME('{self}')"
    @property
    def mime_type(self) -> str:
        "MIME type specification: <type>/<subtype>"
        if ';' in self:
            return self[:self.find(';')]
        return self
    @property
    def type(self) -> str:
        "MIME type"
        return self[:self.find('/')]
    @property
    def subtype(self) -> str:
        "MIME subtype"
        if ';' in self:
            return self[self.find('/')+1:self.find(';')]
        return self[self.find('/')+1]
    @property
    def params(self) -> Dict[str, str]:
        "MIME parameters"
        if ';' in self:
            return {k.strip(): v.strip() for k, v in (x.split('=') for x in self[self.find(';')+1:].split(';'))}
        return {}

class PyExpr(str):
    """Source code for Python expression.

It behaves like `str`, but checks that value is a valid Python expression, and provides
direct access to compiled code.

Raises:
    SyntaxError: When string value is not a valid Python expression.
"""
    _expr_ = None
    def __new__(cls, value: str):
        expr = compile(value, "PyExpr", 'eval')
        new = str.__new__(cls, value)
        new._expr_ = expr
        return new
    def __repr__(self):
        return f"PyExpr('{self}')"
    def get_callable(self, arguments: str='', namespace: Dict[str, Any]=None) -> Callable:
        """Returns expression as callable function ready for execution.

Arguments:
    arguments: String with arguments (names separated by coma) for returned function.
"""
        ns = {}
        if namespace:
            ns.update(namespace)
        code = compile(f"def expr({arguments}):\n    return {self}",
                       "PyExpr", 'exec')
        eval(code, ns)
        return ns['expr']
    @property
    def expr(self):
        "Expression code ready to be appased to `eval`."
        return self._expr_

class PyCode(str):
    """Python source code.

It behaves like `str`, but checks that value is a valid Python code block, and provides
direct access to compiled code.

Raises:
    SyntaxError: When string value is not a valid Python code block.
"""
    _code_ = None
    def __new__(cls, value: str):
        code = compile(value, "PyCode", 'exec')
        new = str.__new__(cls, value)
        new._code_ = code
        return new
    @property
    def code(self):
        "Python code ready to be appased to `exec`."
        return self._code_

class PyCallable(str):
    """Source code for Python callable.

It behaves like `str`, but checks that value is a valid Python callable (function of class
definition), and acts like a callable (i.e. you can directly call the PyCallable value).

Raises:
    ValueError: When string value does not contains the function or class definition.
    SyntaxError: When string value is not a valid Python callable.
"""
    _callable_ = None
    def __new__(cls, value: str):
        callable_name = None
        for line in value.split('\n'):
            if line.lower().startswith('def '):
                callable_name = line[4:line.find('(')].strip()
                break
        if callable_name is None:
            for line in value.split('\n'):
                if line.lower().startswith('class '):
                    callable_name = line[6:line.find('(')].strip()
                    break
        if callable_name is None:
            raise ValueError("Python function or class definition not found")
        ns = {}
        eval(compile(value, "PyCallable", 'exec'), ns)
        new = str.__new__(cls, value)
        new._callable_ = ns[callable_name]
        new.name = callable_name
        return new
    def __call__(self, *args, **kwargs):
        return self._callable_(*args, **kwargs)

# Metaclasses

def Conjunctive(name, bases, attrs):
    """Returns a metaclass that is conjunctive descendant of all metaclasses used by parent
classes.

Example:

    class A(type): pass

    class B(type): pass

    class AA(metaclass=A):pass

    class BB(metaclass=B):pass

    class CC(AA, BB, metaclass=Conjunctive): pass
"""
    basemetaclasses = []
    for base in bases:
        metacls = type(base)
        if isinstance(metacls, type) and metacls is not type and not metacls in basemetaclasses:
            basemetaclasses.append(metacls)
    dynamic = type(''.join(b.__name__ for b in basemetaclasses), tuple(basemetaclasses), {})
    return dynamic(name, bases, attrs)



#coding:utf-8
#
# PROGRAM/MODULE: firebird-base
# FILE:           test/test_protobuf.py
# DESCRIPTION:    Unit tests for firebird.base.protobuf
# CREATED:        21.5.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""firebird-base - Unit tests for firebird.base.protobuf

"""

from __future__ import annotations
import unittest
from enum import IntEnum
from firebird.base.types import Error
from firebird.base.protobuf import register_decriptor, get_enum_type, get_enum_field_type, \
     get_enum_value_name, is_enum_registered, is_msg_registered, create_message, \
     ProtoEnumType, _enumreg, _msgreg
from base_test_pb2 import DESCRIPTOR

ENUM_TYPE_NAME: str = 'firebird.base.TestEnum'
STATE_MSG_TYPE_NAME: str = 'firebird.base.TestState'
COLLECTION_MSG_TYPE_NAME: str = 'firebird.base.TestCollection'


class TestProtobuf(unittest.TestCase):
    """Unit tests for firebird.base.types"""
    def __init__(self, methodName='runTest'):
        super().__init__(methodName)
    def setUp(self) -> None:
        _enumreg.clear()
        _msgreg.clear()
    def tearDown(self):
        pass
    def test_aaa_register(self):
        self.assertEqual(len(_msgreg), 0)
        self.assertEqual(len(_enumreg), 0)
        #
        self.assertFalse(is_enum_registered(ENUM_TYPE_NAME))
        self.assertFalse(is_msg_registered(STATE_MSG_TYPE_NAME))
        self.assertFalse(is_msg_registered(COLLECTION_MSG_TYPE_NAME))
        #
        register_decriptor(DESCRIPTOR)
        #
        self.assertTrue(is_enum_registered(ENUM_TYPE_NAME))
        self.assertTrue(is_msg_registered(STATE_MSG_TYPE_NAME))
        self.assertTrue(is_msg_registered(COLLECTION_MSG_TYPE_NAME))
        #
        self.assertEqual(len(_msgreg), 2)
        self.assertEqual(len(_enumreg), 1)
        self.assertIn(ENUM_TYPE_NAME, _enumreg)
        self.assertIn(STATE_MSG_TYPE_NAME, _msgreg)
        self.assertIn(COLLECTION_MSG_TYPE_NAME, _msgreg)
    def test_enums(self):
        class TestEnum(IntEnum):
            UNKNOWN = 0
            READY = 1
            RUNNING = 2
            WAITING = 3
            SUSPENDED = 4
            FINISHED = 5
            ABORTED = 6
            # Aliases
            CREATED = 1
            BLOCKED = 3
            STOPPED = 4
            TERMINATED = 6

        enum_spec = [('TEST_UNKNOWN', 0),
                     ('TEST_READY', 1),
                     ('TEST_RUNNING', 2),
                     ('TEST_WAITING', 3),
                     ('TEST_SUSPENDED', 4),
                     ('TEST_FINISHED', 5),
                     ('TEST_ABORTED', 6),
                     ('TEST_CREATED', 1),
                     ('TEST_BLOCKED', 3),
                     ('TEST_STOPPED', 4),
                     ('TEST_TERMINATED', 6),
                     ]
        register_decriptor(DESCRIPTOR)
        # Value name
        self.assertEqual(get_enum_value_name(ENUM_TYPE_NAME, TestEnum.SUSPENDED),
                         f'TEST_{TestEnum.SUSPENDED.name}')
        # Errors
        with self.assertRaises(KeyError) as cm:
            get_enum_value_name('BAD.TYPE', TestEnum.SUSPENDED)
        self.assertEqual(cm.exception.args, ("Unregistered protobuf enum type 'BAD.TYPE'",))
        with self.assertRaises(KeyError) as cm:
            get_enum_value_name(ENUM_TYPE_NAME, 9999)
        self.assertEqual(cm.exception.args, (f"Enum {ENUM_TYPE_NAME} has no name defined for value 9999",))
        # Type specification
        enum: ProtoEnumType = get_enum_type(ENUM_TYPE_NAME)
        self.assertEqual(enum.name, ENUM_TYPE_NAME)
        self.assertEqual(enum.get_value_name(TestEnum.SUSPENDED),
                         f'TEST_{TestEnum.SUSPENDED.name}')
        self.assertListEqual(enum.items(), enum_spec)
        self.assertListEqual(enum.keys(), [k for k, v in enum_spec])
        self.assertListEqual(enum.values(), [v for k, v in enum_spec])
        # attribute access to enum values
        for name, value in enum_spec:
            self.assertEqual(getattr(enum, name), value)
        with self.assertRaises(AttributeError) as cm:
            enum.TEST_BAD_VALUE
        self.assertEqual(cm.exception.args, (f"Enum {ENUM_TYPE_NAME} has no value with name 'TEST_BAD_VALUE'",))
    def test_messages(self):
        register_decriptor(DESCRIPTOR)
        #
        msg = create_message(STATE_MSG_TYPE_NAME)
        self.assertIsNotNone(msg)
        self.assertEqual(get_enum_field_type(msg, 'test'), ENUM_TYPE_NAME)
        #
        msg.name = 'State.NAME'
        msg.test = 1
        # Errors
        with self.assertRaises(KeyError) as cm:
            create_message('NOT_REGISTERED')
        self.assertEqual(cm.exception.args, ("Unregistered protobuf message 'NOT_REGISTERED'",))
        with self.assertRaises(KeyError) as cm:
            get_enum_field_type(msg, 'BAD_FIELD')
        self.assertEqual(cm.exception.args, ("Message does not have field 'BAD_FIELD'",))




#coding:utf-8
#
# PROGRAM/MODULE: firebird-base
# FILE:           firebird/base/collections.py
# DESCRIPTION:    Various collection types
# CREATED:        14.5.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2019 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________.

"""Firebird Base - Various collection types

This module provides data structures that behave much like builtin `list` and `dict` types,
but with direct support of operations that can use structured data stored in container, and
which would normally require utilization of `operator`, `functools` or other means.

All containers provide next operations:

* `filter` and `filterfalse` that return generator that yields items for which `expr` is
  evaluated as True (or False).
* `find` that returns first item for which `expr` is evaluated as True, or default.
* `contains` that returns True if there is any item for which `expr` is evaluated as True.
* `occurrence` that returns number of items for which `expr` is evaluated as True.
* `all` and `any` that return True if `expr` is evaluated as True for all or any list element(s).
* `report` that returns generator that yields data produced by expression(s) evaluated on
  list items.

Individual collection types provide additional operations like splitting and extracting
based on expression etc.

Expressions used by these methods could be strings that contain Python expression
referencing the collection item(s), or lambda functions.
"""

from __future__ import annotations
from typing import Type, Union, Any, Dict, List, Tuple, Mapping, Sequence, Generator, \
     Iterable, Callable, cast
from operator import attrgetter
import copy
from .types import Error, Distinct, Sentinel, UNDEFINED


def make_lambda(expr: str, params: str='item', context: Dict=None):
    """Makes lambda function from expression."""
    return eval(f'lambda {params}:{expr}', context) if context else eval(f'lambda {params}:{expr}')


#: Collection Item
Item = Type[Any]
#: Collection Item type specification
TypeSpec = Union[Type, Tuple[Type]]
#: Collection Item sort expression
ItemExpr = Union[str, Callable[[Item], Item]]
#: Filter expression
FilterExpr = Union[str, Callable[[Item], bool]]
#: Check expression
CheckExpr = Union[str, Callable[[Item, Any], bool]]

class BaseObjectCollection:
    """Collection of objects.
"""
    def filter(self, expr: FilterExpr) -> Generator[Item, None, None]:
        """Returns generator that yields items for which `expr` is evaluated as True.

Arguments:
    expr: Bool expression, a callable accepting one parameter and returning bool or
          bool expression as string referencing list item as `item`.

Example:
    .. code-block:: python

       L.filter(lambda x: x.name.startswith("ABC"))
       L.filter('item.name.startswith("ABC")')
"""
        fce = expr if callable(expr) else make_lambda(expr)
        return (item for item in self if fce(item))
    def filterfalse(self, expr: FilterExpr) -> Generator[Item, None, None]:
        """Returns generator that yields items for which `expr` is evaluated as False.

Arguments:
    expr: Bool expression, a callable accepting one parameter and returning bool or
          bool expression as string referencing list item as `item`.

Example:
    .. code-block:: python

       L.filterfalse(lambda x: x.name.startswith("ABC"))
       L.filterfalse('item.name.startswith("ABC")')
"""
        fce = expr if callable(expr) else make_lambda(expr)
        return (item for item in self if not fce(item))
    def find(self, expr: FilterExpr, default: Any=None) -> Item:
        """Returns first item for which `expr` is evaluated as True, or default.

Arguments:
    expr:    Bool expression, a callable accepting one parameter and returning bool or
             bool expression as string referencing list item as `item`.
    default: Default value returned when Items is not found.

Example:
    .. code-block:: python

       L.find(lambda x: x.name.startswith("ABC"))
       L.find('item.name.startswith("ABC")')
"""
        for item in self.filter(expr):
            return item
        return default
    def contains(self, expr: FilterExpr=None) -> bool:
        """Returns True if there is any item for which `expr` is evaluated as True.

Arguments:
    expr: Bool expression, a callable accepting one parameter and returning bool or
          bool expression as string referencing list item as `item`.

Example:
    .. code-block:: python

       if L.contains(lambda x: x.name.startswith("ABC")):
           ...
       if L.contains('item.name.startswith("ABC")'):
           ...
"""
        return self.find(expr) is not None
    def report(self, *args) -> Generator[Any, None, None]:
        """Returns generator that yields data produced by expression(s) evaluated on list items.

Arguments:
    args: Parameter(s) could be one from:

        - A callable accepting one parameter and returning data for output
        - One or more expressions as string referencing item as `item`.

Example:
    .. code-block:: python

       # generator of tuples with item.name and item.size

       L.report(lambda x: (x.name, x.size))
       L.report('item.name','item.size')

       # generator of item names

       L.report(lambda x: x.name)
       L.report('item.name')
"""
        if len(args) == 1 and callable(args[0]):
            fce = args[0]
        else:
            attrs = f'({",".join(args) if len(args) > 1 else args[0]})'
            fce = make_lambda(attrs)
        return (fce(item) for item in self)
    def occurrence(self, expr: FilterExpr) -> int:
        """Return number of items for which `expr` is evaluated as True.

Arguments:
    expr: Bool expression, a callable accepting one parameter and returning bool or
          bool expression as string referencing list item as `item`.

Example:
    .. code-block:: python

       L.occurrence(lambda x: x.name.startswith("ABC"))
       L.occurrence('item.name.startswith("ABC")')
"""
        return sum(1 for item in self.filter(expr))
    def all(self, expr: FilterExpr) -> bool:
        """Return True if `expr` is evaluated as True for all list elements.

Arguments:
    expr: Bool expression, a callable accepting one parameter and returning bool or
          bool expression as string referencing list item as `item`.

Example:
    .. code-block:: python

       L.all(lambda x: x.name.startswith("ABC"))
       L.all('item.name.startswith("ABC")')
"""
        fce = expr if callable(expr) else make_lambda(expr)
        for item in self:
            if not fce(item):
                return False
        return True
    def any(self, expr: FilterExpr) -> bool:
        """Return True if `expr` is evaluated as True for any list element.

Arguments:
    expr: Bool expression, a callable accepting one parameter and returnin bool or
          bool expression as string referencing list item as `item`.

Example:
    .. code-block:: python

       L.any(lambda x: x.name.startswith("ABC"))
       L.any('item.name.startswith("ABC")')
"""
        fce = expr if callable(expr) else make_lambda(expr)
        for item in self:
            if fce(item):
                return True
        return False

class DataList(list, BaseObjectCollection):
    """List of data (objects) with additional functionality.

Arguments:
    items:     Sequence to initialize the collection.
    type_spec: Reject instances that are not instances of specified types.
    key_expr:  Key expression. Must contain item referrence as `item`, for example
               `item.attribute_name`. If **all** classes specified in `type_spec` are
               descendants of `.Distinct`, the default value is `item.get_key()`, otherwise
               the default is `None`.
    frozen:    Create frozen list.

Raises:
    ValueError: When initialization sequence contains invalid instance.
"""
    def __init__(self, items: Iterable=None, type_spec: TypeSpec=UNDEFINED,
                 key_expr: str=None, frozen: bool=False):
        assert key_expr is None or isinstance(key_expr, str)
        assert key_expr is None or make_lambda(key_expr) is not None
        if items is not None:
            super().__init__(items)
        else:
            super().__init__()
        if type_spec is not UNDEFINED and key_expr is None:
            all_distinct = True
            if isinstance(type_spec, tuple):
                for ts in type_spec:
                    all_distinct = all_distinct and issubclass(ts, Distinct)
            else:
                all_distinct = all_distinct and issubclass(type_spec, Distinct)
            if all_distinct:
                key_expr = 'item.get_key()'
        self.__key_expr: str = key_expr
        self.__frozen: bool = False
        self._type_spec: TypeSpec = type_spec
        self.__map: Dict = None
        if frozen:
            self.freeze()
    def __valchk(self, value: Item) -> None:
        if self._type_spec is not UNDEFINED and not isinstance(value, self._type_spec):
            raise TypeError("Value is not an instance of allowed class")
    def __updchk(self) -> None:
        if self.__frozen:
            raise TypeError("Cannot modify frozen DataList")
    def __setitem__(self, index, value) -> None:
        self.__updchk()
        if isinstance(index, slice):
            for val in value:
                self.__valchk(val)
        else:
            self.__valchk(value)
        super().__setitem__(index, value)
    def __delitem__(self, index) -> None:
        self.__updchk()
        super().__delitem__(index)
    def __contains__(self, o):
        if self.__map is not None:
            if isinstance(o, Distinct) and self.__key_expr == 'item.get_key()':
                return o.get_key() in self.__map
            return make_lambda(self.__key_expr)(o) in self.__map
        return super().__contains__(o)
    def insert(self, index: int, item: Item) -> None:
        """Insert item before index.

Raises:
    TypeError: When item is not an instance of allowed class, or list is frozen
"""
        self.__updchk()
        self.__valchk(item)
        super().insert(index, item)
    def remove(self, item: Item) -> None:
        """Remove first occurrence of item.

Raises:
    ValueError: If the value is not present, or list is frozen
"""
        self.__updchk()
        super().remove(item)
    def append(self, item: Item) -> None:
        """Add an item to the end of the list.

Raises:
    TypeError: When item is not an instance of allowed class, or list is frozen
"""
        self.__updchk()
        self.__valchk(item)
        super().append(item)
    def extend(self, iterable: Iterable) -> None:
        """Extend the list by appending all the items in the given iterable.

Raises:
    TypeError: When item is not an instance of allowed class, or list is frozen
"""
        for item in iterable:
            self.append(item)
    def sort(self, attrs: List=None, expr: ItemExpr=None, reverse: bool=False) -> None:
        """Sort items in-place, optionaly using attribute values as key or key expression.

Arguments:
    attrs: List of attribute names.
    expr: Key expression, a callable accepting one parameter or expression
           as string referencing list item as `item`.

    Important:
        Only one parameter (`attrs` or `expr`) could be specified.
        If none is present then uses default list sorting rule.

Example:
    .. code-block:: python

       L.sort(attrs=['name','degree'])       # Sort by item.name, item.degree
       L.sort(expr=lambda x: x.name.upper()) # Sort by upper item.name
       L.sort(expr='item.name.upper()')      # Sort by upper item.name
"""
        assert attrs is None or isinstance(attrs, (list, tuple))
        if attrs:
            super().sort(key=attrgetter(*attrs), reverse=reverse)
        elif expr:
            super().sort(key=expr if callable(expr) else make_lambda(expr), reverse=reverse)
        elif self.__key_expr:
            super().sort(key=make_lambda(self.__key_expr), reverse=reverse)
        else:
            super().sort(reverse=reverse)
    def clear(self) -> None:
        """Remove all items from the list.

Raises:
    TypeError: When list is frozen."""
        self.__updchk()
        super().clear()
    def freeze(self) -> None:
        """Set list to immutable (frozen) state.

Freezing list makes internal map from `key_expr` to item index that significantly
speeds up retrieval by key using the `get()` method.

It's not possible to `add`, `delete` or `change` items in frozen list, but `.sort` is allowed.
"""
        self.__frozen = True
        if self.__key_expr:
            fce = make_lambda(self.__key_expr)
            self.__map = dict(((key, index) for index, key in enumerate((fce(item) for item in self))))
    def split(self, expr: FilterExpr) -> Tuple['DataList', 'DataList']:
        """Return two new `DataList` instances, first with items for which `expr` is
evaluated as True and second for `expr` evaluated as False.

Arguments:
    expr: A callable accepting one parameter or expression as string referencing list
          item as `item`.

Example:
    .. code-block:: python

       below, above = L.split(lambda x: x.size > 100)
       below, above = L.split('item.size > 100')
"""
        return DataList(self.filter(expr), self._type_spec, self.__key_expr), \
               DataList(self.filterfalse(expr), self._type_spec, self.__key_expr)
    def extract(self, expr: FilterExpr):
        """Move items for which `expr` is evaluated as True into new `DataList`.

Arguments:
    expr: A callable accepting one parameter or expression as string referencing list item
          as `item`.

Raises:
    TypeError: When list is frozen

Example:
    .. code-block:: python

       L.extract(lambda x: x.name.startswith("ABC"))
       L.extract('item.name.startswith("ABC")')
"""
        self.__updchk()
        fce = expr if callable(expr) else make_lambda(expr)
        l = DataList(type_spec=self._type_spec, key_expr=self.__key_expr)
        i = 0
        while len(self) > i:
            item = self[i]
            if fce(item):
                l.append(item)
                del self[i]
            else:
                i += 1
        return l
    def get(self, key: Any, default: Any=None) -> Item:
        """Returns item with given key using default key expression. Returns `default`
value if item is not found.

Uses very fast method to look up value of default key expression in `frozen` list,
otherwise it uses slower list traversal.

Arguments:
    key:     Searched value.
    default: Default value.

Raises:
    Error: If key expression is not defined.

Example:
    .. code-block:: python

       # Search using default key expression (fast for frozen list)
       L.get('ITEM_NAME')
"""
        if not self.__key_expr:
            raise Error("Key expression required")
        if self.__map:
            i = self.__map.get(key)
            return default if i is None else self[i]
        fce = make_lambda(f'{self.__key_expr} == key', 'item, key')
        for item in self:
            if fce(item, key):
                return item
        return default
    @property
    def frozen(self) -> bool:
        "True if list items couldn't be changed"
        return self.__frozen
    @property
    def key_expr(self) -> Item:
        "Key expression"
        return self.__key_expr
    @property
    def type_spec(self) -> Union[TypeSpec, Sentinel]:
        "Specification of valid type(s) for list values, or `.UNDEFINED` if there is no such constraint."
        return self._type_spec

class Registry(BaseObjectCollection, Mapping[Any, Distinct]):
    """Mapping container for `.Distinct` objects.

Arguments:
    data: Either a `.Distinct` instance, or sequence or mapping of `.Distinct` instances.

Any method that expects a `key` also acepts `.Distinct` instance.

To store items into registry with existence check, use:
    - R.store(item)
    - R.extend(items) or R.extend(item)

To update items in registry (added if not present), use:
    - R[key] = item
    - R.update(items) or R.update(item)

To check presence of item or key in registry, use:
    - key in R

To retrieve items from registry, use:
    - R.get(key, default=None)
    - R[key]
    - R.pop(key, default=None)
    - R.popitem(last=True)

To delete items from registry, use:
    - R.remove(item)
    - del R[key]

Whenever a `key` is required, you can use either a `Distinct` instance, or any value that
represens a key value for instances of stored type.
"""
    def __init__(self, data: Union[Mapping, Sequence, Registry]=None):
        self._reg: Dict = {}
        if data:
            self.update(data)
    def __len__(self):
        return len(self._reg)
    def __getitem__(self, key):
        return self._reg[key.get_key() if isinstance(key, Distinct) else key]
    def __setitem__(self, key, value):
        assert isinstance(value, Distinct)
        self._reg[key.get_key() if isinstance(key, Distinct) else key] = value
    def __delitem__(self, key):
        del self._reg[key.get_key() if isinstance(key, Distinct) else key]
    def __iter__(self):
        return iter(self._reg.values())
    def __repr__(self):
        return f"{self.__class__.__name__}(" \
               f"[{', '.join(repr(x) for x in self)}])"
    def __contains__(self, item):
        if isinstance(item, Distinct):
            item = item.get_key()
        return item in self._reg
    def clear(self) -> None:
        """Remove all items from registry."""
        self._reg.clear()
    def get(self, key: Any, default: Any=None) -> Distinct:
        """ D.get(key[,d]) -> D[key] if key in D else d. d defaults to None."""
        return self._reg.get(key.get_key() if isinstance(key, Distinct) else key, default)
    def store(self, item: Distinct) -> Distinct:
        """Register an item.

Raises:
    ValueError: When item is already registered.
"""
        assert isinstance(item, Distinct), f"Item is not of type '{Distinct.__name__}'"
        key = item.get_key()
        if key in self._reg:
            raise ValueError(f"Item already registered, key: '{key}'")
        self._reg[key] = item
        return item
    def remove(self, item: Distinct):
        """Removes item from registry (same as: del R[item])."""
        del self._reg[item.get_key()]
    def update(self, _from: Union[Distinct, Mapping, Sequence]) -> None:
        """Update items in the registry.

Arguments:
    _from: Either a `.Distinct` instance, or sequence or mapping of `.Distinct` instances.
"""
        if isinstance(_from, Distinct):
            self[_from] = _from
        else:
            for item in cast(Mapping, _from).values() if hasattr(_from, 'values') else _from:
                self[item] = item
    def extend(self, _from: Union[Distinct, Mapping, Sequence]) -> None:
        """Store one or more items to the registry.

Arguments:
    _from: Either a `.Distinct` instance, or sequence or mapping of `.Distinct` instances.
"""
        if isinstance(_from, Distinct):
            self.store(_from)
        else:
            for item in cast(Mapping, _from).values() if hasattr(_from, 'values') else _from:
                self.store(item)
    def copy(self) -> Registry:
        """Shalow copy of the registry."""
        if self.__class__ is Registry:
            return Registry(self)
        data = self._reg
        try:
            self._reg = {}
            c = copy.copy(self)
        finally:
            self._reg = data
        c.update(self)
        return c
    def pop(self, key: Any, default: Any=None) -> Distinct:
        """Remove specified `key` and return the corresponding `.Distinct` object. If `key`
        is not found, the `default` is returned if given, otherwise `KeyError` is raised """
        return self._reg.pop(key.get_key() if isinstance(key, Distinct) else key, default)
    def popitem(self, last: bool=True) -> Distinct:
        """Returns and removes a `.Distinct` object. The objects are returned in LIFO order
if `last` is true or FIFO order if false."""
        if last:
            _, item = self._reg.popitem()
            return item
        item = next(iter(self))
        self.remove(item)
        return item



#!/usr/bin/python
#coding:utf-8
#
# PROGRAM/MODULE: firebird-base
# FILE:           test/test_trace.py
# DESCRIPTION:    Unit tests for firebird.base.trace
# CREATED:        21.5.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""firebird-base - Unit tests for firebird.base.logging
"""

from __future__ import annotations
import unittest
import os
from logging import getLogger, Formatter, lastResort, LogRecord
from firebird.base.types import *
from firebird.base.logging import LoggingIdMixin, LogLevel
from firebird.base.trace import trace_manager, TraceFlag, add_trace, traced, TracedMixin

class Namespace:
    "Simple Namespace"

DECORATED = Namespace()
DECORATED.name = 'DECORATED'

class BaseLoggingTest(unittest.TestCase):
    "Base class for logging unit tests"
    def __init__(self, methodName='runTest'):
        super().__init__(methodName)
        self._saved_trace: TraceFlag = None
        self.logger = getLogger()
        self.logger.setLevel(LogLevel.NOTSET)
        #self.mngr: LoggingManager = get_manager()
        self.fmt: Formatter = Formatter("%(levelname)10s: [%(name)s] topic='%(topic)s' agent=%(agent)s context=%(context)s %(message)s")
        lastResort.setLevel(LogLevel.NOTSET)
    def setUp(self) -> None:
        self._saved_trace = trace_manager.trace
        trace_manager.clear()
        self.logger.handlers.clear()
        #lastResort.setFormatter(self.fmt)
        #self.logger.addHandler(lastResort)
    def tearDown(self):
        trace_manager.trace = self._saved_trace
        if 'FBASE_TRACE' in os.environ:
            del os.environ['FBASE_TRACE']
    def show(self, records, attrs=None):
        while records:
            item = records.pop(0)
            try:
                print({k: v for k, v in vars(item).items() if attrs is None or k in attrs})
            except:
                print(item)

class Traced(LoggingIdMixin, TracedMixin):
    "traceable callables"
    def __init__(self, owner: BaseLoggingTest, logging_id: str=None):
        self.owner = owner
        if logging_id is not None:
            self._logging_id_ = logging_id
    def traced_noparam_noresult(self) -> None:
        getLogger().info('<traced_noparam_noresult>')
    def traced_noparam_result(self) -> str:
        getLogger().info('<traced_noparam_result>')
        return 'OK'
    def traced_param_noresult(self, pos_only, / , pos, kw='KW', *, kw_only='KW_ONLY') -> None:
        getLogger().info('<traced_param_noresult>')
    def traced_param_result(self, pos_only, / , pos, kw='KW', *, kw_only='KW_ONLY') -> str:
        getLogger().info('<traced_param_result>')
        return 'OK'
    def traced_long_result(self) -> str:
        getLogger().info('<traced_long_result>')
        return '0123456789' * 10
    def traced_raises(self) -> None:
        getLogger().info('<traced_raises>')
        raise Error("No cookies left")

class DecoratedTraced(LoggingIdMixin):
    "traceable callables"
    def __init__(self, owner: BaseLoggingTest, logging_id: str=None):
        self.owner = owner
        if logging_id is not None:
            self._logging_id_ = logging_id
    @traced()
    def traced_noparam_noresult(self) -> None:
        getLogger().info('<traced_noparam_noresult>')
    @traced()
    def traced_noparam_result(self) -> str:
        getLogger().info('<traced_noparam_result>')
        return 'OK'
    @traced()
    def traced_param_noresult(self, pos_only, / , pos, kw='KW', *, kw_only='KW_ONLY') -> None:
        getLogger().info('<traced_param_noresult>')
    @traced()
    def traced_param_result(self, pos_only, / , pos, kw='KW', *, kw_only='KW_ONLY') -> str:
        getLogger().info('<traced_param_result>')
        return 'OK'
    @traced()
    def traced_raises(self) -> None:
        getLogger().info('<traced_raises>')
        raise Error("No cookies left")


class TestTraced(BaseLoggingTest):
    """Unit tests for firebird.base.logging"""
    def setUp(self) -> None:
        super().setUp()
        if not __debug__:
            os.environ['FBASE_TRACE'] = 'on'
        trace_manager.trace |= TraceFlag.ACTIVE
    def verify_func(self, records, func_name: str, only: bool=False) -> None:
        if only:
            self.assertEqual(len(records), 1)
        else:
            self.assertGreaterEqual(len(records), 1)
        self.assertEqual(records.pop(0).message, f'<{func_name}>')
    def test_aaa(self):
        "Default settings only, events: FAIL"

        def verify(records, func_name, params: str='', result: str=None,
                   outcome: str=('log_failed', '<--')) -> None:
            self.assertGreaterEqual(len(records), 2)
            self.verify_func(records, func_name)
            rec = records.pop(0)
            self.assertEqual(rec.name, 'trace')
            self.assertEqual(rec.levelno, LogLevel.DEBUG)
            self.assertEqual(rec.args, ())
            self.assertEqual(rec.filename, 'trace.py')
            self.assertEqual(rec.module, 'trace')
            self.assertEqual(rec.funcName, outcome[0])
            self.assertEqual(rec.topic, 'trace')
            self.assertEqual(rec.agent, 'Traced')
            self.assertEqual(rec.context, UNDEFINED)
            self.assertTrue(rec.message.startswith(f'{outcome[1]} {func_name}'))
            self.assertTrue(rec.message.endswith(f'{result}'))

        self.assertEqual(trace_manager.trace, TraceFlag.ACTIVE | TraceFlag.FAIL)
        ctx = Traced(self)
        # traced_noparam_noresult
        with self.assertLogs(level='DEBUG') as log:
            traced()(ctx.traced_noparam_noresult)()
        self.verify_func(log.records, 'traced_noparam_noresult', True)
        # traced_noparam_result
        with self.assertLogs(level='DEBUG') as log:
            traced()(ctx.traced_noparam_result)()
        self.verify_func(log.records, 'traced_noparam_result', True)
        # traced_param_noresult
        with self.assertLogs(level='DEBUG') as log:
            traced()(ctx.traced_param_noresult)(1, 2, kw_only='NO-DEFAULT')
        self.verify_func(log.records, 'traced_param_noresult', True)
        # traced_param_noresult
        with self.assertLogs(level='DEBUG') as log:
            traced()(ctx.traced_param_result)(1, 2, kw_only='NO-DEFAULT')
        self.verify_func(log.records, 'traced_param_result', True)
        # traced_raises
        with self.assertLogs(level='DEBUG') as log:
            with self.assertRaises(Error):
                traced()(ctx.traced_raises)()
        verify(log.records, 'traced_raises', result='Error: No cookies left')
    def test_aab(self):
        "Default decorator settings, all events"
        def verify(records, func_name: str, params: str='', result: str='',
                   outcome: str=('log_after', '<<<')) -> None:
            self.assertEqual(len(records), 3)
            rec: LogRecord = records.pop(0)
            self.assertEqual(rec.name, 'trace')
            self.assertEqual(rec.levelno, LogLevel.DEBUG)
            self.assertEqual(rec.args, ())
            self.assertEqual(rec.filename, 'trace.py')
            self.assertEqual(rec.module, 'trace')
            self.assertEqual(rec.funcName, 'log_before')
            self.assertEqual(rec.topic, 'trace')
            self.assertEqual(rec.agent, 'Traced')
            self.assertEqual(rec.context, UNDEFINED)
            self.assertEqual(rec.message, f'>>> {func_name}({params})')
            #
            self.verify_func(records, func_name)
            #
            rec = records.pop(0)
            self.assertEqual(rec.name, 'trace')
            self.assertEqual(rec.levelno, LogLevel.DEBUG)
            self.assertEqual(rec.args, ())
            self.assertEqual(rec.filename, 'trace.py')
            self.assertEqual(rec.module, 'trace')
            self.assertEqual(rec.funcName, outcome[0])
            self.assertEqual(rec.topic, 'trace')
            self.assertEqual(rec.agent, 'Traced')
            self.assertEqual(rec.context, UNDEFINED)
            self.assertTrue(rec.message.startswith(f'{outcome[1]} {func_name}'))
            self.assertTrue(rec.message.endswith(f'{result}'))

        ctx = Traced(self)
        trace_manager.trace |= (TraceFlag.FAIL | TraceFlag.BEFORE | TraceFlag.AFTER)

        # traced_noparam_noresult
        with self.assertLogs(level='DEBUG') as log:
            traced()(ctx.traced_noparam_noresult)()
        verify(log.records, 'traced_noparam_noresult')
        # traced_noparam_result
        with self.assertLogs(level='DEBUG') as log:
            traced()(ctx.traced_noparam_result)()
        verify(log.records, 'traced_noparam_result', result='OK')
        # traced_param_noresult
        with self.assertLogs(level='DEBUG') as log:
            traced()(ctx.traced_param_noresult)(1, 2, kw_only='NO-DEFAULT')
        verify(log.records, 'traced_param_noresult', "pos_only=1, pos=2, kw='KW', kw_only='NO-DEFAULT'")
        # traced_param_noresult
        with self.assertLogs(level='DEBUG') as log:
            traced()(ctx.traced_param_result)(1, 2, kw_only='NO-DEFAULT')
        verify(log.records, 'traced_param_result', "pos_only=1, pos=2, kw='KW', kw_only='NO-DEFAULT'", 'OK')
        # traced_raises
        with self.assertLogs(level='DEBUG') as log:
            with self.assertRaises(Error):
                traced()(ctx.traced_raises)()
        verify(log.records, 'traced_raises', result='Error: No cookies left', outcome=('log_failed', '<--'))
    def test_custom_msg(self):
        def verify(records, msg_before: str, msg_after_start: str, msg_after_end: str='') -> None:
            self.assertEqual(len(records), 3)
            rec = records.pop(0)
            self.assertEqual(rec.message, msg_before)
            records.pop(0)
            rec = records.pop(0)
            self.assertTrue(rec.message.startswith(msg_after_start))
            self.assertTrue(rec.message.endswith(msg_after_end))

        ctx = Traced(self)
        trace_manager.trace |= (TraceFlag.FAIL | TraceFlag.BEFORE | TraceFlag.AFTER)
        #
        with self.assertLogs(level='DEBUG') as log:
            d = traced(msg_before='ENTER {_fname_} ({pos_only}, {pos}, {kw}, {kw_only})')
            d(ctx.traced_param_noresult)(1, 2, kw_only='NO-DEFAULT')
        verify(log.records, 'ENTER traced_param_noresult (1, 2, KW, NO-DEFAULT)', '<<< traced_param_noresult')
        with self.assertLogs(level='DEBUG') as log:
            d = traced(msg_after='EXIT {_fname_}: {_result_}')
            d(ctx.traced_param_noresult)(1, 2, kw_only='NO-DEFAULT')
        verify(log.records, ">>> traced_param_noresult(pos_only=1, pos=2, kw='KW', kw_only='NO-DEFAULT')",
               'EXIT traced_param_noresult: None', '')
        d = traced(msg_before='ENTER {_fname_} ({pos_only}, {pos}, {kw}, {kw_only})',
                   msg_after='EXIT {_fname_}: {_result_}')
        with self.assertLogs(level='DEBUG') as log:
            d(ctx.traced_param_noresult)(1, 2, kw_only='NO-DEFAULT')
        verify(log.records, 'ENTER traced_param_noresult (1, 2, KW, NO-DEFAULT)',
               'EXIT traced_param_noresult: None', '')
        with self.assertLogs(level='DEBUG') as log:
            d = traced(msg_before='ENTER {_fname_} ()',
                       msg_after='EXIT {_fname_}: {_result_}',
                       msg_failed='!!! {_fname_}: {_exc_}')
            with self.assertRaises(Error):
                d(ctx.traced_raises)()
        verify(log.records, 'ENTER traced_raises ()',
               '!!! traced_raises: Error: No cookies left', '')
    def test_extra(self):
        def foo(bar=''):
            return f'Foo{bar}!'

        def verify(records, msg_before: str, msg_after: str) -> None:
            self.assertEqual(len(records), 3)
            self.assertEqual(records.pop(0).message, msg_before)
            records.pop(0)
            self.assertEqual(records.pop(0).message, msg_after)

        ctx = Traced(self)
        trace_manager.trace |= (TraceFlag.FAIL | TraceFlag.BEFORE | TraceFlag.AFTER)
        #
        with self.assertLogs(level='DEBUG') as log:
            d = traced(msg_before='>>> {_fname_} ({foo()}, {foo(kw)}, {foo(kw_only)})',
                       msg_after='<<< {_fname_}: {foo(_result_)}', extra={'foo': foo})
            d(ctx.traced_param_noresult)(1, 2, kw_only='bar')
        verify(log.records, '>>> traced_param_noresult (Foo!, FooKW!, Foobar!)',
               '<<< traced_param_noresult: FooNone!')
    def test_topic(self):
        def verify(records, topic: str) -> None:
            self.assertEqual(len(records), 3)
            self.assertEqual(records.pop(0).topic, topic)
            records.pop(0)
            self.assertEqual(records.pop(0).topic, topic)

        ctx = Traced(self)
        trace_manager.trace |= (TraceFlag.FAIL | TraceFlag.BEFORE | TraceFlag.AFTER)
        #
        with self.assertLogs(level='DEBUG') as log:
            traced(topic='fun')(ctx.traced_noparam_noresult)()
        verify(log.records, 'fun')
    def test_max_param_length(self):
        def verify(records, message: str, result: str='Result: OK') -> None:
            self.assertEqual(len(records), 3)
            self.assertEqual(records.pop(0).message, message)
            records.pop(0)
            self.assertTrue(records.pop(0).message.endswith(result))

        ctx = Traced(self)
        trace_manager.trace |= (TraceFlag.FAIL | TraceFlag.BEFORE | TraceFlag.AFTER)
        #
        with self.assertLogs(level='DEBUG') as log:
            traced(max_param_length=10)(ctx.traced_param_result)('123456789', '0123456789' * 10)
        verify(log.records, ">>> traced_param_result(pos_only='123456789', pos='0123456789..[90]', kw='KW', kw_only='KW_ONLY')")
        #
        with self.assertLogs(level='DEBUG') as log:
            traced(max_param_length=10)(ctx.traced_long_result)()
        verify(log.records, '>>> traced_long_result()', 'Result: 0123456789..[90]')
    def test_agent_ctx(self):
        def verify(records, agent, ctx) -> None:
            self.assertEqual(len(records), 3)
            rec = records.pop(0)
            self.assertEqual(rec.agent, agent)
            self.assertEqual(rec.context, ctx)
            records.pop(0)
            rec = records.pop(0)
            self.assertEqual(rec.agent, agent)
            self.assertEqual(rec.context, ctx)

        ctx = Traced(self)
        trace_manager.trace |= (TraceFlag.FAIL | TraceFlag.BEFORE | TraceFlag.AFTER)
        #
        with self.assertLogs(level='DEBUG') as log:
            traced(agent=UNDEFINED, context=UNDEFINED)(ctx.traced_noparam_noresult)()
        verify(log.records, UNDEFINED, UNDEFINED)
        ctx.log_context = '<CONTEXT>'
        ctx._logging_id_ = '<AGENT>'
        with self.assertLogs(level='DEBUG') as log:
            traced()(ctx.traced_noparam_noresult)()
        verify(log.records, '<AGENT>', '<CONTEXT>')
    def test_level(self):
        def verify(records, level) -> None:
            self.assertEqual(len(records), 3)
            self.assertEqual(records.pop(0).levelno, level)
            records.pop(0)
            self.assertEqual(records.pop(0).levelno, level)

        ctx = Traced(self)
        trace_manager.trace |= (TraceFlag.FAIL | TraceFlag.BEFORE | TraceFlag.AFTER)
        #
        with self.assertLogs(level='DEBUG') as log:
            traced(level=LogLevel.INFO)(ctx.traced_noparam_noresult)()
        verify(log.records, LogLevel.INFO)
    def test_forced(self):
        def verify(records, msg_before: str, msg_after_start: str, msg_after_end: str='') -> None:
            self.assertEqual(len(records), 3)
            rec = records.pop(0)
            self.assertEqual(rec.message, msg_before)
            records.pop(0)
            rec = records.pop(0)
            self.assertTrue(rec.message.startswith(msg_after_start))
            self.assertTrue(rec.message.endswith(msg_after_end))

        ctx = Traced(self)
        trace_manager.trace = (TraceFlag.FAIL | TraceFlag.BEFORE | TraceFlag.AFTER)
        with self.assertLogs(level='DEBUG') as log:
            traced()(ctx.traced_noparam_noresult)()
        self.verify_func(log.records, 'traced_noparam_noresult', True)
        #
        with self.assertLogs(level='DEBUG') as log:
            traced(flags=TraceFlag.ACTIVE)(ctx.traced_noparam_noresult)()
        verify(log.records, '>>> traced_noparam_noresult()', '<<< traced_noparam_noresult')
    def test_env(self):
        ctx = Traced(self)
        trace_manager.trace = (TraceFlag.ACTIVE | TraceFlag.FAIL | TraceFlag.BEFORE | TraceFlag.AFTER)
        with self.assertLogs(level='DEBUG') as log:
            traced()(ctx.traced_noparam_noresult)()
        self.assertEqual(len(log.records), 3)
        os.environ['FBASE_TRACE'] = 'off'
        with self.assertLogs(level='DEBUG') as log:
            traced()(ctx.traced_noparam_noresult)()
        self.verify_func(log.records, 'traced_noparam_noresult', True)
        del os.environ['FBASE_TRACE']
    def test_debug(self):
        if __debug__:
            self.skipTest("__debug__ is True")
        if 'FBASE_TRACE' in os.environ:
            del os.environ['FBASE_TRACE']
        ctx = Traced(self)
        trace_manager.trace = (TraceFlag.ACTIVE | TraceFlag.FAIL | TraceFlag.BEFORE | TraceFlag.AFTER)
        with self.assertLogs(level='DEBUG') as log:
            traced()(ctx.traced_noparam_noresult)()
        self.verify_func(log.records, 'traced_noparam_noresult', True)
    def test_decorated(self):
        "Default settings only, events: FAIL"

        def verify(records, func_name, params: str='', result: str=None,
                   outcome: str=('log_failed', '<--')) -> None:
            self.assertGreaterEqual(len(records), 2 if __debug__ else 1)
            self.verify_func(records, func_name)
            if __debug__:
                rec = records.pop(0)
                self.assertEqual(rec.name, 'trace')
                self.assertEqual(rec.levelno, LogLevel.DEBUG)
                self.assertEqual(rec.args, ())
                self.assertEqual(rec.filename, 'trace.py')
                self.assertEqual(rec.module, 'trace')
                self.assertEqual(rec.funcName, outcome[0])
                self.assertEqual(rec.topic, 'trace')
                self.assertEqual(rec.agent, 'DecoratedTraced')
                self.assertEqual(rec.context, UNDEFINED)
                self.assertTrue(rec.message.startswith(f'{outcome[1]} {func_name}'))
                self.assertTrue(rec.message.endswith(f'{result}'))

        self.assertEqual(trace_manager.trace, TraceFlag.ACTIVE | TraceFlag.FAIL)
        ctx = DecoratedTraced(self)
        # traced_noparam_noresult
        with self.assertLogs(level='DEBUG') as log:
            ctx.traced_noparam_noresult()
        self.verify_func(log.records, 'traced_noparam_noresult', True)
        # traced_noparam_result
        with self.assertLogs(level='DEBUG') as log:
            ctx.traced_noparam_result()
        self.verify_func(log.records, 'traced_noparam_result', True)
        # traced_param_noresult
        with self.assertLogs(level='DEBUG') as log:
            ctx.traced_param_noresult(1, 2, kw_only='NO-DEFAULT')
        self.verify_func(log.records, 'traced_param_noresult', True)
        # traced_param_noresult
        with self.assertLogs(level='DEBUG') as log:
            ctx.traced_param_result(1, 2, kw_only='NO-DEFAULT')
        self.verify_func(log.records, 'traced_param_result', True)
        # traced_raises
        with self.assertLogs(level='DEBUG') as log:
            with self.assertRaises(Error):
                ctx.traced_raises()
        verify(log.records, 'traced_raises', result='Error: No cookies left')
    def test_add_traced(self):
        "Default settings only, events: FAIL"

        def verify(records, func_name, params: str='', result: str=None,
                   outcome: str=('log_failed', '<--')) -> None:
            self.assertGreaterEqual(len(records), 2)
            self.verify_func(records, func_name)
            rec = records.pop(0)
            self.assertEqual(rec.name, 'trace')
            self.assertEqual(rec.levelno, LogLevel.DEBUG)
            self.assertEqual(rec.args, ())
            self.assertEqual(rec.filename, 'trace.py')
            self.assertEqual(rec.module, 'trace')
            self.assertEqual(rec.funcName, outcome[0])
            self.assertEqual(rec.topic, 'trace')
            self.assertEqual(rec.agent, 'Traced')
            self.assertEqual(rec.context, UNDEFINED)
            self.assertTrue(rec.message.startswith(f'{outcome[1]} {func_name}'))
            self.assertTrue(rec.message.endswith(f'{result}'))

        self.assertEqual(trace_manager.trace, TraceFlag.ACTIVE | TraceFlag.FAIL)
        add_trace(Traced, 'traced_noparam_noresult')
        add_trace(Traced, 'traced_noparam_result')
        add_trace(Traced, 'traced_param_noresult')
        add_trace(Traced, 'traced_param_result')
        add_trace(Traced, 'traced_raises')
        ctx = Traced(self)
        # traced_noparam_noresult
        with self.assertLogs(level='DEBUG') as log:
            ctx.traced_noparam_noresult()
        self.verify_func(log.records, 'traced_noparam_noresult', True)
        # traced_noparam_result
        with self.assertLogs(level='DEBUG') as log:
            ctx.traced_noparam_result()
        self.verify_func(log.records, 'traced_noparam_result', True)
        # traced_param_noresult
        with self.assertLogs(level='DEBUG') as log:
            ctx.traced_param_noresult(1, 2, kw_only='NO-DEFAULT')
        self.verify_func(log.records, 'traced_param_noresult', True)
        # traced_param_result
        with self.assertLogs(level='DEBUG') as log:
            ctx.traced_param_result(1, 2, kw_only='NO-DEFAULT')
        self.verify_func(log.records, 'traced_param_result', True)
        # traced_raises
        with self.assertLogs(level='DEBUG') as log:
            with self.assertRaises(Error):
                traced()(ctx.traced_raises)()
        verify(log.records, 'traced_raises', result='Error: No cookies left')


if __name__ == '__main__':
    unittest.main()



#coding:utf-8
#
# PROGRAM/MODULE: firebird-base
# FILE:           firebird/base/strconv.py
# DESCRIPTION:    Data conversion from/to string
# CREATED:        4.6.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""firebird-base - Data conversion from/to string
"""

from __future__ import annotations
from typing import Hashable, Callable, Any, Type, Union
from dataclasses import dataclass
from decimal import Decimal, DecimalException
from enum import Enum, IntEnum, IntFlag
from uuid import UUID
from .types import Distinct, MIME, ZMQAddress
from .collections import Registry

#: Function that converts typed value to its string representation.
TConvertToStr = Callable[[Any], str]
#: Function that converts string representation of typed value to typed value.
TConvertFromStr = Callable[[Type, str], Any]

@dataclass
class Convertor(Distinct):
    "Data convertor registry entry."
    cls: Type
    to_str: TConvertToStr
    from_str: TConvertFromStr
    def get_key(self) -> Hashable:
        """Returns instance key."""
        return self.cls
    @property
    def name(self) -> str:
        "Type name"
        return self.cls.__name__
    @property
    def full_name(self) -> str:
        "Type name incl. source module"
        return f'{self.cls.__module__}.{self.cls.__name__}'

_convertors: Registry = Registry()
_classes = {}

# Convertors

#: Valid string literals for True value.
TRUE_STR = ['yes', 'true', 'on', 'y', '1']
#: Valid string literals for False value.
FALSE_STR = ['no', 'false', 'off', 'n', '0']

def any2str(value: Any) -> str:
    "Converts value to string using `str(value)`."
    return str(value)

def str2any(cls: Type, value: str) -> Any:
    "Converts string to data type value using `type(value)`."
    return cls(value)

def register_convertor(cls: Type, *,
                       to_str: TConvertToStr=any2str,
                       from_str: TConvertFromStr=str2any):
    """Registers convertor function(s).

Arguments:
    cls:      Class or class name
    to_str:   Function that converts `cls` value to `str`
    from_str: Function that converts `str` to value of `cls` data type
"""
    _convertors.store(Convertor(cls, to_str, from_str))

def register_class(cls: Type) -> None:
    """Registers class for name lookup.

.. seealso:: `has_convertor()`, `get_convertor()`

Raises:
    TypeError: When class name is already registered.
"""
    if cls.__name__ in _classes:
        raise TypeError(f"Class '{cls.__name__}' already registered as '{_classes[cls.__name__]!r}'")
    _classes[cls.__name__] = cls

def _get_convertor(cls: Union[Type, str]) -> Convertor:
    if isinstance(cls, str):
        if cls in _classes:
            cls = _classes[cls]
    if isinstance(cls, str):
        conv = list(_convertors.filter(f"item.{'full_name' if '.' in cls else 'name'} == '{cls}'"))
        conv = conv.pop(0) if conv else None
    else:
        if (conv := _convertors.get(cls)) is None:
            for base in cls.__mro__:
                conv = _convertors.get(base)
                if conv is not None:
                    break
    return conv

def has_convertor(cls: Union[Type, str]) -> bool:
    """Returns True if class has a convertor.

Arguments:
    cls: Type or type name. The name could be simple class name, or full name that includes
         the module name.

Note:
    When `cls` is a name:

    1. If class name is NOT registered via `register_class()`, it's not possible to perform
       lookup for bases classes.
    2. If simple class name is provided and multiple classes of the same name but from
       different modules have registered convertors, the first one found is used. If you
       want to avoid this situation, use full names.
"""
    return _get_convertor(cls) is not None

def update_convertor(cls: Union[Type, str], *,
                     to_str: TConvertToStr=None,
                     from_str: TConvertFromStr=None):
    """Update convertor function(s).

Arguments:
    cls:      Class or class name
    to_str:   Function that converts `cls` value to `str`
    from_str: Function that converts `str` to value of `cls` data type

Raises:
    KeyError: If data type has not registered convertor.
"""
    conv = get_convertor(cls)
    if to_str:
        conv.to_str = to_str
    if from_str:
        conv.from_str = from_str

def convert_to_str(value: Any) -> str:
    """Converts value to string using registered convertor.

Arguments:
    value:  Value to be converted.

If there is no convertor for value's class, uses MRO to locate alternative convertor.

Raises:
    TypeError: If there is no convertor for value's class or any from its bases classes.
"""
    return get_convertor(value.__class__).to_str(value)


def convert_from_str(cls: Union[Type, str], value: str) -> Any:
    """Converts value from string to data type using registered convertor.

Arguments:
    cls:   Type or type name. The name could be simple class name, or full name that includes
           the module name.
    value: String value to be converted

Note:
    When `cls` is a type name:

    1. If class name is NOT registered via `register_class()`, it's not possible to perform
       lookup for bases classes.
    2. If simple class name is provided and multiple classes of the same name but from
       different modules have registered convertors, the first one found is used. If you
       want to avoid this situation, use full names.

Raises:
    TypeError: If there is no convertor for `cls` or any from its bases classes.
"""
    return get_convertor(cls).from_str(cls, value)

def get_convertor(cls: Union[Type, str]) -> Convertor:
    """Returns Convertor for data type.

Arguments:
    cls: Type or type name. The name could be simple class name, or full name that includes
         the module name.

Note:
    When `cls` is a type name:

    1. If class name is NOT registered via `register_class()`, it's not possible to perform
       lookup for bases classes.
    2. If simple class name is provided and multiple classes of the same name but from
       different modules have registered convertors, the first one found is used. If you
       want to avoid this situation, use full names.

Raises:
    TypeError: If there is no convertor for `cls` or any from its bases classes.
"""
    if (conv := _get_convertor(cls)) is None:
        raise TypeError(f"Type '{cls.__name__ if isinstance(cls, type) else cls}' has no Convertor")
    return conv

def register():

    def bool2str(value: bool) -> str:
        return TRUE_STR[0] if value else FALSE_STR[0]
    def str2bool(type_: Type, value: str) -> bool:
        if (v := value.lower()) in TRUE_STR:
            return True
        if v not in FALSE_STR:
            raise ValueError("Value is not a valid bool string constant")
        return False
    def str2decimal(type_: Type, value: str) -> Decimal:
        try:
            return type_(value)
        except DecimalException:
            raise ValueError(f"could not convert string to {type_.__name__}: '{value}'")
    def enum2str(value: Enum) -> str:
        "Converts any Enum/Flag value to string"
        return value.name
    def str2enum(cls: Type, value: str) -> Enum:
        "Converts string to Enum/Flag value"
        return {k.lower(): v for k, v in cls.__members__.items()}[value.lower()]

    register_convertor(str)
    register_convertor(int)
    register_convertor(float)
    register_convertor(complex)
    register_convertor(Decimal, from_str=str2decimal)
    register_convertor(UUID)
    register_convertor(MIME)
    register_convertor(ZMQAddress)
    register_convertor(bool, to_str=bool2str, from_str=str2bool)
    register_convertor(Enum, to_str=enum2str, from_str=str2enum)
    # We must register IntEnum and IntFlag because 'int' is before Enum in MRO
    register_convertor(IntEnum, to_str=enum2str, from_str=str2enum)
    register_convertor(IntFlag, to_str=enum2str, from_str=str2enum)

register()
del register



#!/usr/bin/python
#coding:utf-8
#
# PROGRAM/MODULE: firebird-base
# FILE:           test/test_logging.py
# DESCRIPTION:    Unit tests for firebird.base.logging
# CREATED:        21.5.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""firebird-base - Unit tests for firebird.base.logging
"""

from __future__ import annotations
import unittest
from logging import getLogger, Formatter, lastResort, LogRecord
from firebird.base.types import *
from firebird.base.logging import logging_manager, get_logger, bind_logger, \
     LogLevel, BindFlag, install_null_logger

class Namespace:
    "Simple Namespace"

DECORATED = Namespace()
DECORATED.name = 'DECORATED'

class BaseLoggingTest(unittest.TestCase):
    "Base class for logging unit tests"
    def __init__(self, methodName='runTest'):
        super().__init__(methodName)
        self.logger = getLogger()
        self.logger.setLevel(LogLevel.NOTSET)
        self.fmt: Formatter = Formatter("%(levelname)10s: [%(name)s] topic='%(topic)s' agent=%(agent)s context=%(context)s %(message)s")
        lastResort.setLevel(LogLevel.NOTSET)
    def setUp(self) -> None:
        logging_manager.clear()
        self.logger.handlers.clear()
        #lastResort.setFormatter(self.fmt)
        #self.logger.addHandler(lastResort)
    def tearDown(self):
        pass
    def show(self, records, attrs=None):
        while records:
            item = records.pop(0)
            try:
                print({k: v for k, v in vars(item).items() if attrs is None or k in attrs})
            except:
                print(item)

class TestLogging(BaseLoggingTest):
    """Unit tests for firebird.base.logging"""
    def test_module(self):
        self.assertIsNotNone(lastResort)
        self.assertIsNotNone(lastResort.formatter)
    def test_aaa(self):
        # root
        with self.assertLogs() as log:
            get_logger(self).info('Message')
        rec: LogRecord = log.records.pop(0)
        self.assertEqual(rec.name, 'root')
        self.assertEqual(rec.levelno, LogLevel.INFO)
        self.assertEqual(rec.args, ())
        self.assertEqual(rec.module, 'test_logging')
        self.assertEqual(rec.filename, 'test_logging.py')
        self.assertEqual(rec.funcName, 'test_aaa')
        self.assertEqual(rec.topic, '')
        self.assertEqual(rec.agent, 'test_aaa (test_logging.TestLogging)')
        self.assertEqual(rec.context, UNDEFINED)
        self.assertEqual(rec.message, 'Message')
        # trace
        with self.assertLogs() as log:
            get_logger(self, topic='trace').info('Message')
        rec = log.records.pop(0)
        self.assertEqual(rec.name, 'trace')
        self.assertEqual(rec.levelno, LogLevel.INFO)
        self.assertEqual(rec.args, ())
        self.assertEqual(rec.module, 'test_logging')
        self.assertEqual(rec.filename, 'test_logging.py')
        self.assertEqual(rec.funcName, 'test_aaa')
        self.assertEqual(rec.topic, 'trace')
        self.assertEqual(rec.agent, 'test_aaa (test_logging.TestLogging)')
        self.assertEqual(rec.context, UNDEFINED)
        self.assertEqual(rec.message, 'Message')
    def test_interpolation(self):
        data = ['interpolation', 'breakdown', 'overflow']
        # Using keyword arguments
        with self.assertLogs() as log:
            get_logger(self).info('Information {data}', data=data[0])
        rec: LogRecord = log.records.pop(0)
        self.assertEqual(rec.message, 'Information interpolation')
        # Using positional dictionary
        with self.assertLogs() as log:
            get_logger(self).info('Information {data}', {'data': data[1]})
        rec = log.records.pop(0)
        self.assertEqual(rec.message, 'Information breakdown')
        # Using positional args
        with self.assertLogs() as log:
            get_logger(self).info('Information {args[0][2]}', data)
        rec = log.records.pop(0)
        self.assertEqual(rec.message, 'Information overflow')
    def test_bind(self):
        LOG_AC = 'test.agent.ctx'
        LOG_AX = 'test.agent.ANY'
        LOG_XC = 'test.ANY.ctx'
        LOG_XX = 'test.ANY.ANY'
        ctx = Namespace()
        ctx.logging_id = 'CONTEXT'
        ctx_B = Namespace()
        ctx_B.logging_id = 'B-CONTEXT'
        agent = Namespace()
        agent.logging_id = 'AGENT'
        agent_B = Namespace()
        agent_B.logging_id = 'B-AGENT'
        #
        logging_manager.bind_logger(agent, ctx, LOG_AC)
        logging_manager.bind_logger(agent, ANY, LOG_AX)
        logging_manager.bind_logger(ANY, ctx, LOG_XC)
        # root logger unmasked
        with self.assertLogs(level='DEBUG') as log:
            get_logger(agent_B, ctx_B).debug('General:B-Agent:B-context')
        rec: LogRecord = log.records.pop(0)
        self.assertEqual(rec.name, 'root')
        # this will mask the root logger, we also test the `bind_logger()`
        bind_logger(ANY, ANY, LOG_XX)
        #
        self.assertTrue(BindFlag.ANY_AGENT in logging_manager.bindings)
        self.assertTrue(BindFlag.ANY_CTX in logging_manager.bindings)
        self.assertTrue(BindFlag.ANY_ANY in logging_manager.bindings)
        self.assertTrue(BindFlag.DIRECT in logging_manager.bindings)
        self.assertEqual(len(logging_manager.loggers), 4)
        self.assertEqual(len(logging_manager.topics), 1)
        self.assertEqual(logging_manager.topics[''], 4)
        #
        with self.assertLogs(level='DEBUG') as log:
            get_logger(agent, ctx).debug('General:Agent:Context')
        rec = log.records.pop(0)
        self.assertEqual(rec.name, LOG_AC)
        #
        with self.assertLogs(level='DEBUG') as log:
            get_logger(agent, ctx_B).debug('General:Agent:B-context')
        rec = log.records.pop(0)
        self.assertEqual(rec.name, LOG_AX)
        #
        with self.assertLogs(level='DEBUG') as log:
            get_logger(agent, ANY).debug('General:Agent:ANY')
        rec = log.records.pop(0)
        self.assertEqual(rec.name, LOG_AX)
        #
        with self.assertLogs(level='DEBUG') as log:
            get_logger(agent).debug('General:Agent:UNDEFINED')
        rec = log.records.pop(0)
        self.assertEqual(rec.name, LOG_AX)
        #
        with self.assertLogs(level='DEBUG') as log:
            get_logger(agent_B, ctx).debug('General:B-Agent:Context')
        rec = log.records.pop(0)
        self.assertEqual(rec.name, LOG_XC)
        #
        with self.assertLogs(level='DEBUG') as log:
            get_logger(context=ctx).debug('General:UNDEFINED:Context')
        rec = log.records.pop(0)
        self.assertEqual(rec.name, LOG_XC)
        #
        with self.assertLogs(level='DEBUG') as log:
            get_logger(agent_B, ctx_B).debug('General:B-Agent:B-context')
        rec = log.records.pop(0)
        self.assertEqual(rec.name, LOG_XX)
        #
        with self.assertLogs(level='DEBUG') as log:
            get_logger().debug('General:UNDEFINED:UNDEFINED')
        rec = log.records.pop(0)
        self.assertEqual(rec.name, LOG_XX)
    def test_clear(self):
        LOG_AC = 'test.agent.ctx'
        LOG_AX = 'test.agent.ANY'
        LOG_XC = 'test.ANY.ctx'
        LOG_XX = 'test.ANY.ANY'
        ctx = Namespace()
        ctx.logging_id = 'CONTEXT'
        agent = Namespace()
        agent.logging_id = 'AGENT'
        #
        logging_manager.bind_logger(agent, ctx, LOG_AC)
        logging_manager.bind_logger(agent, ANY, LOG_AX)
        logging_manager.bind_logger(ANY, ctx, LOG_XC)
        logging_manager.bind_logger(ANY, ANY, LOG_XX)
        #
        self.assertTrue(BindFlag.ANY_AGENT in logging_manager.bindings)
        self.assertTrue(BindFlag.ANY_CTX in logging_manager.bindings)
        self.assertTrue(BindFlag.ANY_ANY in logging_manager.bindings)
        self.assertTrue(BindFlag.DIRECT in logging_manager.bindings)
        self.assertEqual(len(logging_manager.loggers), 4)
        self.assertEqual(len(logging_manager.topics), 1)
        # Clear
        logging_manager.clear()
        self.assertFalse(BindFlag.ANY_AGENT in logging_manager.bindings)
        self.assertFalse(BindFlag.ANY_CTX in logging_manager.bindings)
        self.assertFalse(BindFlag.ANY_ANY in logging_manager.bindings)
        self.assertFalse(BindFlag.DIRECT in logging_manager.bindings)
        self.assertEqual(len(logging_manager.loggers), 0)
        self.assertEqual(len(logging_manager.topics), 0)
    def test_unbind(self):
        LOG_AC = 'test.agent.ctx'
        LOG_AX = 'test.agent.ANY'
        LOG_XC = 'test.ANY.ctx'
        LOG_XX = 'test.ANY.ANY'
        ctx = Namespace()
        ctx.logging_id = 'CONTEXT'
        ctx_B = Namespace()
        ctx_B.logging_id = 'B-CONTEXT'
        agent = Namespace()
        agent.logging_id = 'AGENT'
        agent_B = Namespace()
        agent_B.logging_id = 'B-AGENT'
        #
        logging_manager.bind_logger(agent, ctx, LOG_AC)
        logging_manager.bind_logger(agent, ANY, LOG_AX)
        logging_manager.bind_logger(ANY, ctx, LOG_XC)
        logging_manager.bind_logger(ANY, ANY, LOG_XX)
        #
        self.assertTrue(BindFlag.ANY_AGENT in logging_manager.bindings)
        self.assertTrue(BindFlag.ANY_CTX in logging_manager.bindings)
        self.assertTrue(BindFlag.ANY_ANY in logging_manager.bindings)
        self.assertTrue(BindFlag.DIRECT in logging_manager.bindings)
        self.assertEqual(len(logging_manager.loggers), 4)
        self.assertEqual(len(logging_manager.topics), 1)
        self.assertEqual(logging_manager.topics[''], 4)
        # Unbind
        # nothing to remove
        self.assertEqual(0, logging_manager.unbind(agent, ctx, 'trace'))
        self.assertEqual(0, logging_manager.unbind(agent, ctx_B))
        self.assertEqual(0, logging_manager.unbind(agent_B, ctx))
        # targeted
        self.assertEqual(1, logging_manager.unbind(ANY, ANY))
        self.assertTrue(BindFlag.ANY_AGENT in logging_manager.bindings)
        self.assertTrue(BindFlag.ANY_CTX in logging_manager.bindings)
        self.assertFalse(BindFlag.ANY_ANY in logging_manager.bindings)
        self.assertTrue(BindFlag.DIRECT in logging_manager.bindings)
        self.assertEqual(len(logging_manager.loggers), 3)
        self.assertEqual(logging_manager.topics[''], 3)
        # group (all agents for context)
        self.assertEqual(2, logging_manager.unbind(ALL, ctx))
        self.assertFalse(BindFlag.ANY_AGENT in logging_manager.bindings)
        self.assertTrue(BindFlag.ANY_CTX in logging_manager.bindings)
        self.assertFalse(BindFlag.ANY_ANY in logging_manager.bindings)
        self.assertFalse(BindFlag.DIRECT in logging_manager.bindings)
        self.assertEqual(len(logging_manager.loggers), 1)
        self.assertEqual(logging_manager.topics[''], 1)
        # rebind
        logging_manager.bind_logger(agent, ctx, LOG_AC)
        logging_manager.bind_logger(agent, ANY, LOG_AX)
        logging_manager.bind_logger(ANY, ctx, LOG_XC)
        logging_manager.bind_logger(ANY, ANY, LOG_XX)
        # group (all contexts for agent)
        self.assertEqual(2, logging_manager.unbind(agent, ALL))
        self.assertTrue(BindFlag.ANY_AGENT in logging_manager.bindings)
        self.assertTrue(BindFlag.ANY_CTX in logging_manager.bindings)
        self.assertTrue(BindFlag.ANY_ANY in logging_manager.bindings)
        self.assertFalse(BindFlag.DIRECT in logging_manager.bindings)
        self.assertEqual(len(logging_manager.loggers), 2)
        self.assertEqual(logging_manager.topics[''], 2)
    def test_null_logger(self):
        with self.assertLogs() as log:
            get_logger(self).info('Message')
        rec: LogRecord = log.records.pop(0)
        self.assertEqual(rec.message, 'Message')
        install_null_logger()
        bind_logger(ANY, ANY, 'null')
        with self.assertRaises(AssertionError) as cm:
            with self.assertLogs() as log:
                get_logger(self).info('Message')
        self.assertEqual(cm.exception.args, ("no logs of level INFO or higher triggered on root",))

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: firebird/base/config.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='firebird/base/config.proto',
  package='firebird.base',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=b'\n\x1a\x66irebird/base/config.proto\x12\rfirebird.base\x1a\x19google/protobuf/any.proto\"\xf0\x01\n\x05Value\x12\x13\n\tas_string\x18\x02 \x01(\tH\x00\x12\x12\n\x08\x61s_bytes\x18\x03 \x01(\x0cH\x00\x12\x11\n\x07\x61s_bool\x18\x04 \x01(\x08H\x00\x12\x13\n\tas_double\x18\x05 \x01(\x01H\x00\x12\x12\n\x08\x61s_float\x18\x06 \x01(\x02H\x00\x12\x13\n\tas_sint32\x18\x07 \x01(\x11H\x00\x12\x13\n\tas_sint64\x18\x08 \x01(\x12H\x00\x12\x13\n\tas_uint32\x18\t \x01(\rH\x00\x12\x13\n\tas_uint64\x18\n \x01(\x04H\x00\x12&\n\x06\x61s_msg\x18\x0b \x01(\x0b\x32\x14.google.protobuf.AnyH\x00\x42\x06\n\x04kind\"\x93\x02\n\x0b\x43onfigProto\x12\x38\n\x07options\x18\x01 \x03(\x0b\x32\'.firebird.base.ConfigProto.OptionsEntry\x12\x38\n\x07\x63onfigs\x18\x02 \x03(\x0b\x32\'.firebird.base.ConfigProto.ConfigsEntry\x1a\x44\n\x0cOptionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12#\n\x05value\x18\x02 \x01(\x0b\x32\x14.firebird.base.Value:\x02\x38\x01\x1aJ\n\x0c\x43onfigsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12)\n\x05value\x18\x02 \x01(\x0b\x32\x1a.firebird.base.ConfigProto:\x02\x38\x01\x62\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_any__pb2.DESCRIPTOR,])




_VALUE = _descriptor.Descriptor(
  name='Value',
  full_name='firebird.base.Value',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='as_string', full_name='firebird.base.Value.as_string', index=0,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='as_bytes', full_name='firebird.base.Value.as_bytes', index=1,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='as_bool', full_name='firebird.base.Value.as_bool', index=2,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='as_double', full_name='firebird.base.Value.as_double', index=3,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='as_float', full_name='firebird.base.Value.as_float', index=4,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='as_sint32', full_name='firebird.base.Value.as_sint32', index=5,
      number=7, type=17, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='as_sint64', full_name='firebird.base.Value.as_sint64', index=6,
      number=8, type=18, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='as_uint32', full_name='firebird.base.Value.as_uint32', index=7,
      number=9, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='as_uint64', full_name='firebird.base.Value.as_uint64', index=8,
      number=10, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='as_msg', full_name='firebird.base.Value.as_msg', index=9,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='kind', full_name='firebird.base.Value.kind',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=73,
  serialized_end=313,
)


_CONFIGPROTO_OPTIONSENTRY = _descriptor.Descriptor(
  name='OptionsEntry',
  full_name='firebird.base.ConfigProto.OptionsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='firebird.base.ConfigProto.OptionsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='firebird.base.ConfigProto.OptionsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=447,
  serialized_end=515,
)

_CONFIGPROTO_CONFIGSENTRY = _descriptor.Descriptor(
  name='ConfigsEntry',
  full_name='firebird.base.ConfigProto.ConfigsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='firebird.base.ConfigProto.ConfigsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='firebird.base.ConfigProto.ConfigsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=517,
  serialized_end=591,
)

_CONFIGPROTO = _descriptor.Descriptor(
  name='ConfigProto',
  full_name='firebird.base.ConfigProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='options', full_name='firebird.base.ConfigProto.options', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='configs', full_name='firebird.base.ConfigProto.configs', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_CONFIGPROTO_OPTIONSENTRY, _CONFIGPROTO_CONFIGSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=316,
  serialized_end=591,
)

_VALUE.fields_by_name['as_msg'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_VALUE.oneofs_by_name['kind'].fields.append(
  _VALUE.fields_by_name['as_string'])
_VALUE.fields_by_name['as_string'].containing_oneof = _VALUE.oneofs_by_name['kind']
_VALUE.oneofs_by_name['kind'].fields.append(
  _VALUE.fields_by_name['as_bytes'])
_VALUE.fields_by_name['as_bytes'].containing_oneof = _VALUE.oneofs_by_name['kind']
_VALUE.oneofs_by_name['kind'].fields.append(
  _VALUE.fields_by_name['as_bool'])
_VALUE.fields_by_name['as_bool'].containing_oneof = _VALUE.oneofs_by_name['kind']
_VALUE.oneofs_by_name['kind'].fields.append(
  _VALUE.fields_by_name['as_double'])
_VALUE.fields_by_name['as_double'].containing_oneof = _VALUE.oneofs_by_name['kind']
_VALUE.oneofs_by_name['kind'].fields.append(
  _VALUE.fields_by_name['as_float'])
_VALUE.fields_by_name['as_float'].containing_oneof = _VALUE.oneofs_by_name['kind']
_VALUE.oneofs_by_name['kind'].fields.append(
  _VALUE.fields_by_name['as_sint32'])
_VALUE.fields_by_name['as_sint32'].containing_oneof = _VALUE.oneofs_by_name['kind']
_VALUE.oneofs_by_name['kind'].fields.append(
  _VALUE.fields_by_name['as_sint64'])
_VALUE.fields_by_name['as_sint64'].containing_oneof = _VALUE.oneofs_by_name['kind']
_VALUE.oneofs_by_name['kind'].fields.append(
  _VALUE.fields_by_name['as_uint32'])
_VALUE.fields_by_name['as_uint32'].containing_oneof = _VALUE.oneofs_by_name['kind']
_VALUE.oneofs_by_name['kind'].fields.append(
  _VALUE.fields_by_name['as_uint64'])
_VALUE.fields_by_name['as_uint64'].containing_oneof = _VALUE.oneofs_by_name['kind']
_VALUE.oneofs_by_name['kind'].fields.append(
  _VALUE.fields_by_name['as_msg'])
_VALUE.fields_by_name['as_msg'].containing_oneof = _VALUE.oneofs_by_name['kind']
_CONFIGPROTO_OPTIONSENTRY.fields_by_name['value'].message_type = _VALUE
_CONFIGPROTO_OPTIONSENTRY.containing_type = _CONFIGPROTO
_CONFIGPROTO_CONFIGSENTRY.fields_by_name['value'].message_type = _CONFIGPROTO
_CONFIGPROTO_CONFIGSENTRY.containing_type = _CONFIGPROTO
_CONFIGPROTO.fields_by_name['options'].message_type = _CONFIGPROTO_OPTIONSENTRY
_CONFIGPROTO.fields_by_name['configs'].message_type = _CONFIGPROTO_CONFIGSENTRY
DESCRIPTOR.message_types_by_name['Value'] = _VALUE
DESCRIPTOR.message_types_by_name['ConfigProto'] = _CONFIGPROTO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Value = _reflection.GeneratedProtocolMessageType('Value', (_message.Message,), {
  'DESCRIPTOR' : _VALUE,
  '__module__' : 'firebird.base.config_pb2'
  # @@protoc_insertion_point(class_scope:firebird.base.Value)
  })
_sym_db.RegisterMessage(Value)

ConfigProto = _reflection.GeneratedProtocolMessageType('ConfigProto', (_message.Message,), {

  'OptionsEntry' : _reflection.GeneratedProtocolMessageType('OptionsEntry', (_message.Message,), {
    'DESCRIPTOR' : _CONFIGPROTO_OPTIONSENTRY,
    '__module__' : 'firebird.base.config_pb2'
    # @@protoc_insertion_point(class_scope:firebird.base.ConfigProto.OptionsEntry)
    })
  ,

  'ConfigsEntry' : _reflection.GeneratedProtocolMessageType('ConfigsEntry', (_message.Message,), {
    'DESCRIPTOR' : _CONFIGPROTO_CONFIGSENTRY,
    '__module__' : 'firebird.base.config_pb2'
    # @@protoc_insertion_point(class_scope:firebird.base.ConfigProto.ConfigsEntry)
    })
  ,
  'DESCRIPTOR' : _CONFIGPROTO,
  '__module__' : 'firebird.base.config_pb2'
  # @@protoc_insertion_point(class_scope:firebird.base.ConfigProto)
  })
_sym_db.RegisterMessage(ConfigProto)
_sym_db.RegisterMessage(ConfigProto.OptionsEntry)
_sym_db.RegisterMessage(ConfigProto.ConfigsEntry)


_CONFIGPROTO_OPTIONSENTRY._options = None
_CONFIGPROTO_CONFIGSENTRY._options = None
# @@protoc_insertion_point(module_scope)
